/* 
 *  rocks/1of2.c
 *
 *  1 connection from 2 symmetric connection attempts.
 *
 *  Copyright (C) 2001 Victor Zandy
 *  See COPYING for distribution terms.
 */
/*
  Each end simultaneously attempts to establish a new connection with
  its peer.  The problem is getting both ends to agree on one
  connection.  The protocol is as follows:

  Each end sends its peer the address of a socket on which it is
  listening for an incoming connection.  Then it performs a
  non-blocking connect to its peer's listening socket and polls this
  active socket and the passive listening socket, waiting for an
  connection to be established.

  The remaining behavior depends on whether the role of the end in the
  original connection was CLIENT or SERVER.

  SERVERs send an ack byte on the first established connection.  If
  two connections are ready at the same time (i.e., select returns
  them both being ready), then the server first tries the active one.

  CLIENTs wait for an ack byte on established connections.  If both
  are ready, they check the passive one before the active one.

  The first connection on which the read or write is successful is
  chosen as the connection; the other, if established, is closed.

  I am quite interested in making this protocol role independent.  */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>

#include "rs.h"
#include "log.h"

/* return -1 if A is a smaller time value than B,
           0 if they are equal, or
	   1 if A is a bigger time value than B. */
static int
tv_cmp(const struct timeval *a,	const struct timeval *b)
{
	if (a->tv_sec < b->tv_sec)
		return -1;
	if (a->tv_sec > b->tv_sec)
		return 1;
	if (a->tv_usec < b->tv_usec)
		return -1;
	if (a->tv_usec > b->tv_usec)
		return 1;
	return 0;
}

static int
timeout_expired(struct timeval *tout)
{
	struct timeval tv;
	int rv;
	rv = pi_gettimeofday(&tv, NULL);
	assert(!rv);
	return tv_cmp(&tv, tout) >= 0;
}

struct throttle {
	struct timeval last;
	struct timeval interval;
};

static void
check_throttle(struct throttle *th)
{
	int rv;
	struct timeval tv;

retry:
	pi_gettimeofday(&tv, NULL);
	rs_tv_diff(&tv, &th->last, &tv);
	if (tv_cmp(&tv, &th->interval) < 0) {
		rs_tv_diff(&th->interval, &tv, &tv);
		rv = pi_sitfor(&tv);
		if (0 > rv && errno == EINTR)
			goto retry;
		if (0 > rv)
			assert(0);
	}
	pi_gettimeofday(&th->last, NULL);
}

/* IF *PASSIVEP or *ACTIVEP is non-negative, then
   it must a socket for an established connection */
static socket_t
_1of2_client(socket_t *passivep, socket_t *activep)
{
	socket_t pick = INVALID_SOCKET;
	char byte;
	socket_t passive = *passivep;
	socket_t active = *activep;
	int max;
	fd_set fds;
	struct timeval tv;

retry:
	rs_log("1of2: clients's pick: p=%d, a=%d", *passivep, *activep);
	max = 0;
	FD_ZERO(&fds);
	if (passive != INVALID_SOCKET) {
		FD_SET(passive, &fds);
		max = MAX(max, (int)passive);
	}
	if (active != INVALID_SOCKET) {
		FD_SET(active, &fds);
		max = MAX(max, (int)active);
	}
	tv.tv_sec = tv.tv_usec = 0;
	if (0 > select(max+1, &fds, NULL, NULL, &tv)) {
		if (EINTR == errno)
			goto retry;
		else {
			rs_log("1of2: client error");
			return -1;
		}
	}
	if (active != INVALID_SOCKET && FD_ISSET(active, &fds)) {
		if (0 >= rs_xread(active, &byte, sizeof(byte), 0)) {
			pi_closesocket(active);
			*activep = -1;
		} else {
			pick = active;
			pi_closesocket(passive);
		}
	} 
	if (pick == INVALID_SOCKET && passive != INVALID_SOCKET
		&& FD_ISSET(passive, &fds)) {

		if (0 >= rs_xread(passive, &byte, sizeof(byte), 0)) {
			pi_closesocket(passive);
			*passivep = -1;
		} else {
			pick = passive;
			pi_closesocket(active);
		}
	}
	rs_log("1of2: client picked %d", pick);
	return pick;
}

/* IF *PASSIVEP or *ACTIVEP is non-negative, then
   it must a socket for an established connection */
static socket_t
_1of2_server(socket_t *passivep, socket_t *activep)
{
	socket_t pick = INVALID_SOCKET;
	char byte;
	socket_t passive = *passivep;
	socket_t active = *activep;

	rs_log("1of2: server picking: p=%d, a=%d", *passivep, *activep);
	if (active != INVALID_SOCKET) {
		if (0 > rs_xwrite(active, &byte, sizeof(byte))) {
			pi_closesocket(active);
			*activep = -1;
		} else {
			pick = active;
			pi_closesocket(passive);
		}
	} 
	if (pick == INVALID_SOCKET && passive != INVALID_SOCKET) {
		if (0 > rs_xwrite(passive, &byte, sizeof(byte))) {
			pi_closesocket(passive);
			*passivep = -1;
		} else {
			pick = passive;
			pi_closesocket(active);
		}
	}
	rs_log("1of2: server picked %d", pick);
	return pick;
}

/* Bind and return a new socket to ADDR */
static socket_t
xbind(struct sockaddr_in *addr)
{
	socket_t fd;
	int rv;

	fd = socket(AF_INET, SOCK_STREAM, 0);
	if (0 > fd)
		return INVALID_SOCKET;
	if (0 > rs_reuseaddr(fd)) {
		pi_closesocket(fd);
		fd = INVALID_SOCKET;
	}
	rv = bind(fd, (struct sockaddr *) addr, sizeof(*addr));
	if (0 > rv) {
		pi_closesocket(fd);
		rs_log("1of2 bind failed (%s)", strerror(errno));
		return INVALID_SOCKET;
	}
	/* Non blocking avoids timing problem described in
	   Stevens, Network Programming 1, 15.6 */
	if (0 > rs_nonblock(fd, 1)) {
		pi_closesocket(fd);
		return INVALID_SOCKET;
	}
	return fd;
}

/* Start listening on already bound socket FD */
static socket_t
xlisten(socket_t fd)
{
	int rv;
	rv = listen(fd, 1);
	if (0 > rv) {
		pi_closesocket(fd);
		fd = INVALID_SOCKET;
	}
	return fd;
}

/* Return a new socket making a non-blocking connect to ADDR */
static socket_t
xconnect(struct sockaddr_in *addr)
{
	socket_t fd;
	int rv;
	fd = socket(AF_INET, SOCK_STREAM, 0);
	if (0 > fd)
		return INVALID_SOCKET;
	if (0 > rs_nonblock(fd, 1)) {
		pi_closesocket(fd);
		fd = INVALID_SOCKET;
	}
	rv = connect(fd, (struct sockaddr *) addr, sizeof(*addr));
	if (0 > rv && EINPROGRESS != pi_socket_errno()
		&& EWOULDBLOCK != pi_socket_errno()) {

		pi_closesocket(fd);
		fd = INVALID_SOCKET;
	}
	return fd;
}

static int
self_connected(socket_t sd)
{
	int rv;
	struct sockaddr_in locl, peer;
	salen_t len;

	len = sizeof(locl);
	rv = getsockname(sd, (struct sockaddr *) &locl, &len);
	if (0 > rv)
		return -1;
	rv = getpeername(sd, (struct sockaddr *) &peer, &len);
	if (0 > rv)
		return -1;
	return (locl.sin_addr.s_addr == peer.sin_addr.s_addr
		&& locl.sin_port == peer.sin_port);
}

/*  LOCL: local address on which we listen for incoming
          connections
    PEER: peer's listening address
    LS:   if non-negative, TCP socket bound to LOCL (not listening)
    LIM:  maximum time to spend connection attempt; NULL for no timeout
    ROLE: RS_ROLE_CLIENT or RS_ROLE_SERVER
    Return fd of new connection or -1 on if timeout expires.
*/
socket_t
rs_1of2(struct sockaddr_in *locl, struct sockaddr_in *peer, 
	socket_t ls, struct timeval *lim, rs_role role)
{
	int rv, len;
	socket_t ss;          /* connected passive socket, obtained from ls */
	socket_t ns, cs;      /* active sockets; ns is attempting to connect,
                                             cs is connected */
	socket_t pick;             /* Selected connection descriptor */ 
	struct sockaddr_in addr;
	struct throttle th;

	cs = ss = ns = INVALID_SOCKET;
	pick = INVALID_SOCKET;

	if (ls != INVALID_SOCKET)
		ls = xlisten(ls);

	th.interval.tv_sec = 1;
	th.interval.tv_usec = 0;
	pi_gettimeofday(&th.last, NULL);
	while (pick == INVALID_SOCKET) {
		fd_set rfds, wfds;
		struct timeval tv;
		socket_t max;

		if (lim && timeout_expired(lim)) {
			/* timeout */
			pi_closesocket(ls);
			pi_closesocket(ss);
			pi_closesocket(ns);
			return INVALID_SOCKET;
		}

		if (INVALID_SOCKET == cs && INVALID_SOCKET == ss)
			check_throttle(&th);

		if (INVALID_SOCKET == ss && INVALID_SOCKET == ls) {
			ls = xbind(locl);
			if (ls != INVALID_SOCKET)
				ls = xlisten(ls);
		}
		if (INVALID_SOCKET == cs && INVALID_SOCKET == ns)
			ns = xconnect(peer);

		FD_ZERO(&rfds);
		FD_ZERO(&wfds);
		max = 0;

		/* for clients or servers: poll for established connections */
		if (INVALID_SOCKET == cs && ns != INVALID_SOCKET) {
			FD_SET(ns, &wfds);
			FD_SET(ns, &rfds);
			max = MAX(max, ns);
		}
		if (INVALID_SOCKET == ss && ls != INVALID_SOCKET) {
			FD_SET(ls, &rfds);
			max = MAX(max, ls);
		}

		/* for clients: poll for data on connected sockets */
		if (cs != INVALID_SOCKET && RS_ROLE_CLIENT == role) {
			FD_SET(cs, &rfds);
			max = MAX(max, cs);
		}
		if (ss != INVALID_SOCKET && RS_ROLE_CLIENT == role) {
			FD_SET(ss, &rfds);
			max = MAX(max, ss);
		}

		/* imprecise - we could exceed
		   reconnect timeout by the granularity
		   of this timeout; but the reconnect
		   timeout is usually on order of hours */
		tv.tv_sec = 1;
		tv.tv_usec = 0;
		rv = select((int)max + 1, &rfds, &wfds, NULL, &tv);
		if (0 > rv && EINTR == errno)
			continue;   /* interrupt */
		if (0 > rv) {
			rs_log("select failure in 1of2");
			assert(0);
		}
		if (0 == rv) {
			/* timeout */
			/* retry connect; maybe we have a new IP address */
			if (INVALID_SOCKET == cs && ns != INVALID_SOCKET) {
				pi_closesocket(ns);
				ns = -1;
			}
			continue;
		}

		len = sizeof(addr);
		/* check for newly established connections */
		if (INVALID_SOCKET == ss && ls != INVALID_SOCKET
			&& FD_ISSET(ls, &rfds)) {

			ss = accept(ls, (struct sockaddr *) &addr, &len);
			if (ss != INVALID_SOCKET && RS_ROLE_SERVER == role)
				pick = _1of2_server(&ss, &cs);
			continue;
		}
		if (INVALID_SOCKET == cs && ns != INVALID_SOCKET
		    && (FD_ISSET(ns, &wfds) || FD_ISSET(ns, &rfds))) {
			if (0 > getpeername(ns, (struct sockaddr *) &addr,
					    &len)) {
				pi_closesocket(ns);
				ns = INVALID_SOCKET; /* connect failed */
			} else if (self_connected(ns)) {
				pi_closesocket(ns);
				ns = INVALID_SOCKET;
			} else {
				cs = ns;
				ns = INVALID_SOCKET;
				if (RS_ROLE_SERVER == role)
					pick = _1of2_server(&ss, &cs);
			}
		}

		/* One of CS or SS is ready for reading */
		if (RS_ROLE_CLIENT == role
			&& (ss != INVALID_SOCKET || cs != INVALID_SOCKET))

			pick = _1of2_client(&ss, &cs);

	}
	if (pick != INVALID_SOCKET)
		rs_nonblock(pick, 0);
	return pick;
}
