/*
 *  rocks/events_win.c
 *
 *  Handling for network events in windows rocks.
 *
 *  Copyright (C) 2001 Victor Zandy, Greg Quinn
 *  See COPYING for distribution terms.
 */

#include "rs.h"
#include "rs_win.h"
#include "rslist.h"
#include "log.h"
#include <process.h>
#include <assert.h>

#define WM_ROCKS (WM_USER + 1)
#define NUM_ROCKS_MSGS 50

struct event_data {
	SOCKET name;
	SOCKET sd;
	unsigned long net_events;
	int is_async;
	union {
		struct {
			HWND async_wnd;
			unsigned async_msg;
			unsigned rocks_msg; 
		};
		HANDLE ev_event;
	};
	rs_t rs;
	struct event_data *next;
};
struct event_data *event_data_head = NULL;

static HWND event_window;
static HANDLE net_event_thread;

static struct event_data *get_event_data(SOCKET sd) {

	struct event_data *ed = event_data_head;
	while (ed)
		if (ed->sd == sd)
			return ed;
		else
			ed = ed->next;

	return NULL;
}

/* for debugging */
void event_show_message(LPMSG msg) {

	struct event_data *ed = event_data_head;
	while (ed)
		if (ed->async_wnd == msg->hwnd && ed->async_msg == msg->message) {
			/*rs_log("app recv msg: name=%d lParam=0x%x wParam=0x%x",
				ed->name, msg->lParam, msg->wParam);*/
			return;
		}
		else
			ed = ed->next;
}

/* posts messages for all events flagged in ev */
static void post_message(const struct event_data *ed, LPARAM lParam) {

	PostMessage(ed->async_wnd, ed->async_msg, ed->name, lParam);
	//rs_log("posting: name=%d lParam=0x%x", ed->name, lParam);
}

/* message handling for rocks - returns 1 to allow event, 0 to block */
static int rocks_event_handler(struct event_data *ed, LPARAM lParam) {

	long ev = WSAGETSELECTEVENT(lParam);

	switch (ev) {

		case FD_READ:
			return ed->rs->state == RS_ESTABLISHED;

		case FD_WRITE:
			return ed->rs->state == RS_ESTABLISHED;

		case FD_OOB:
			return ed->rs->state == RS_ESTABLISHED;

		case FD_CLOSE:
			rs_reconnect(ed->rs, RS_NOBLOCK);
			return 0;
	
		case FD_CONNECT:
			if (ed->rs->state == RS_ESTABLISHED)
				return 0;
			else if (ed->rs->role != RS_ROLE_CLIENT) {
				rs_continue_connect(ed->rs);
				return 0;
			}
			else {
				rs_init_connection(ed->rs);
				return 1;
			}

		default:
			assert(0);
			return 0;
	}
}

static void handle_event(struct event_data *ed, LPARAM lParam) {

	if (ed->rs) {
		/* exit without sending if rocks_event handler "blocks" event */
		if (!rocks_event_handler(ed, lParam))
			return;
	}

	if (ed->is_async) {
		post_message(ed, lParam);
	}
	else
		assert(0);
}

LRESULT CALLBACK window_proc(HWND wnd, UINT msg, WPARAM wParam, LPARAM lParam) {

	if (msg >= WM_ROCKS && msg < WM_ROCKS + NUM_ROCKS_MSGS) {
		struct event_data *ed;
		/*rs_log("lib recv msg: msg=%u sd=%d ev=0x%x err=0x%x",
		       msg, wParam,
		       WSAGETSELECTEVENT(lParam), WSAGETSELECTERROR(lParam));*/
		ed = get_event_data((SOCKET)wParam);
		if (ed && msg == ed->rocks_msg)
			handle_event(ed, lParam);
		return 0;
	}
	else
		return DefWindowProc(wnd, msg, wParam, lParam);
}

DWORD __stdcall net_event_thread_func(void *null) {

	MSG msg;

	while (GetMessage(&msg, NULL, 0, 0)) {
		TranslateMessage(&msg);
		DispatchMessage(&msg);
	}

	return (DWORD)msg.wParam;
}

int init_net_events() {

	WNDCLASS window_class;

	/* window / message queue stuff */
	window_class.style = CS_HREDRAW | CS_VREDRAW;
	window_class.lpfnWndProc = window_proc;
	window_class.cbClsExtra = 0;
	window_class.cbWndExtra = 0;
	window_class.hInstance = rocks_inst;
	window_class.hIcon = LoadIcon(NULL, IDI_APPLICATION);
	window_class.hCursor = LoadCursor(NULL, IDC_ARROW);
	window_class.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	window_class.lpszMenuName = NULL;
	window_class.lpszClassName = "EventWindow";
	assert(RegisterClassA(&window_class));

	event_window = CreateWindowA("EventWindow", "Event Window",
		                         WS_OVERLAPPEDWINDOW,
	                             CW_USEDEFAULT, CW_USEDEFAULT,
						         CW_USEDEFAULT, CW_USEDEFAULT,
						         NULL, NULL,
						         rocks_inst, NULL);
	assert(event_window != NULL);

	net_event_thread = (HANDLE)_beginthreadex(NULL, 0, net_event_thread_func,
	                                          NULL, 0, NULL);

	return 0;
}

/* returns 1 if rocks found, 0 if not */
int event_unrock(rs_t rs) {

	struct event_data *ed = get_event_data(rs->sd);
	if (ed) {
		ed->rs = NULL;
		return 1;
	}

	return 0;
}

/* replacement for WSAAsyncSelect - sends messages if necessary */
static int async_select(SOCKET s, HWND wnd, unsigned msg, long ev,
						struct event_data *ed) {

	if (ed->rs && ed->rs->state == RS_SUSPENDED)
		return 0;

	assert(WSAAsyncSelect(s, wnd, msg, ev) == 0);

	if ((ev & FD_READ) && ed->rs && (ed->rs->edpspill || ed->rs->clospill))
		post_message(ed, FD_READ);

	return 0;
}

/* remakes the select call when needed */
/* makes assumption that window messages for old socket are to be discarded */
int event_rebind(rs_t rs) {

	struct event_data *ed = event_data_head;

	while (ed)
		if (ed->rs == rs) {
			ed->rocks_msg = (ed->rocks_msg - WM_ROCKS + 1) % NUM_ROCKS_MSGS + WM_ROCKS;
			async_select(rs->sd, event_window, ed->rocks_msg, ed->net_events, ed);
			/*rs_log("event rebind: name=%d sd=%d msg=%u", ed->name, rs->sd,
				ed->rocks_msg);*/
			ed->sd = rs->sd;
			return 1;
		}
		else
			ed = ed->next;

	return 0;
}

/* called when a socket is WSAAsyncSelect()ed */
int add_async_select(SOCKET name, HWND wnd, unsigned msg, long ev, rs_t rs) {

	SOCKET sd;
	struct event_data *ed;
	int alloc = 0;

	sd = rs_lookup_socket(name);
	ed = get_event_data(sd);
	if (!ed) {
		assert(ed = (struct event_data *)malloc(sizeof(struct event_data)));
		alloc = 1;
	}

	ed->name = name;
	ed->sd = sd;
	ed->net_events = ev;
	ed->is_async = 1;
	ed->async_wnd = wnd;
	ed->async_msg = msg;
	ed->rs = rs;

	if (alloc) {
		ed->rocks_msg = WM_ROCKS;
		ed->next = event_data_head;
		event_data_head = ed;
	}

	if (rs)
		rs->nonblock = 1;

	ed->rocks_msg = (ed->rocks_msg - WM_ROCKS + 1) % NUM_ROCKS_MSGS + WM_ROCKS;
	async_select(sd, event_window, ed->rocks_msg, ev, ed);
	/*rs_log("AsyncSelect: name=%d ev=0x%x msg=%u", name, ev, ed->rocks_msg);*/

	return 0;
}

#if 0
int add_event_select(SOCKET name, HANDLE app_event, long ev, rs_t rs) {

	SOCKET sd;
	HANDLE *h_ptr;
	struct event_data *d_ptr;

	get_lock();

	get_data_pointers(name, &h_ptr, &d_ptr);
	sd = rs_lookup_socket(name);
	assert(WSAEventSelect(sd, *h_ptr, ev) == 0);

	d_ptr->name = name;
	d_ptr->sd = sd;
	d_ptr->net_events = ev;
	d_ptr->is_async = 0;
	d_ptr->ev_event = app_event;
	d_ptr->rs = rs;

	release_lock();

	return 0;
}
#endif

int remove_event(SOCKET name) {

	struct event_data *ed = event_data_head;
	struct event_data **link = &event_data_head;

	while (ed)
		if (ed->name == name) {
			*link = ed->next;
			free(ed);
			return 1;
		}
		else {
			link = &ed->next;
			ed = ed->next;
		}

	return 0;
}