/* 
 *  rocks/iop.c
 *
 *  Interoperability of rocks with ordinary sockets.
 *
 *  Copyright (C) 2001 Victor Zandy
 *  See COPYING for distribution terms.
 */
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>

#include "rs.h"
#include "log.h"
#include "rslist.h"		/* for rebind_rock */

#define ROCKIDLEN 16
static char rockid[ROCKIDLEN] = "IROCKYOUROCKMAN";

static char *
xmemmem(char *buf, int buflen,
	char *pat, int patlen)
{
	int i;
	for (i = 0; i < buflen-patlen+1; i++)
		if (!strncmp(&buf[i], pat, patlen))
			return &buf[i];
	return NULL;
}

/* return 1 if peer is enhanced; return 0 if not or if there was an error */
static int
client_probe(rs_t rs)
{
	char buf[ROCKIDLEN];
	int rv;
	ring_t ring;
	fd_set fds;
	struct timeval tv;
	int ret = 0;
	int n;

	shutdown(rs->sd, 1);

	ring = rs_new_ring(100*1024);
	if (!ring)
		return 0;

	while (1) {
		FD_ZERO(&fds);
		FD_SET(rs->sd, &fds);
		tv.tv_sec = 0;
		tv.tv_usec = 500000;
		n = select((int)rs->sd+1, &fds, NULL, NULL, &tv);
		if (0 > n && errno == EINTR)
			continue;
		if (0 > n) {
			rs_log("edp: rock <%p> client probe error", rs);
			break;
		}
		if (0 == n) {
			rs_log("edp: rock <%p> client probe timeout", rs);
			break;
		}
		rv = recv(rs->sd, buf, sizeof(buf), 0);
		if (0 > rv && errno == EINTR)
			continue;
		if (!rv || (0 > rv && errno == ECONNRESET)) {
			/* server closed or reset connection */
			char *p = rs_ring_data(ring);
			int n = rs_ring_nbytes(ring);
			if (n < ROCKIDLEN)
				break;
			if (xmemmem(p, n, rockid, ROCKIDLEN))
				ret = 1;
			break;
		}
		rs_push_ring(ring, buf, rv);
	}
	pi_closesocket(rs->sd);
	rs_free_ring(ring);
	return ret;
}

static unsigned long
room(ring_t ring)
{
	return rs_ring_size(ring) - rs_ring_nbytes(ring);
}

static ring_t
double_ring(ring_t ring)
{
	ring_t r;
	unsigned long sz;
	sz = 2 * rs_ring_size(ring);
	r = rs_new_ring(sz);
	if (!r)
		return NULL;
	rs_push_ring(r, rs_ring_data(ring), rs_ring_nbytes(ring));
	rs_free_ring(ring);
	return r;
}

/* return k, an index into s such that
   &s[k] is the longest suffix of s that
   is a prefix of pat.
   k >= slen means the empty string.
*/
static unsigned long
prematch(char *pat, char *s, int plen, int slen)
{
	unsigned long k, n;

	if (slen > plen) {
		k = slen-plen;
		n = plen;
	} else {
		k = 0;
		n = slen;
	}
	/* n is min(len(&s[k]), len(pat))
	   note that neither s nor pat is necessarily null-terminated */
	while (0 != strncmp(pat, &s[k], n)) {
		k++;
		n--;
	}
	return k;
}

/* read from rs->fd until the rock announcement is seen.
   store any bytes preceeding the announcement in rs->edpspill.
*/
static int
scan_for_rockid(rs_t rs)
{
	ring_t spill;
	char buf[ROCKIDLEN];
	unsigned long k;
	int rv, m;

	spill = rs_new_ring(2048); /* arbitrary starting size */
	if (!spill)
		goto err;

	m = 0;
	while (1) {
		rv = recv(rs->sd, buf+m, ROCKIDLEN-m, 0);
		if (0 > rv && errno == EINTR)
			continue;
		if (0 >= rv)
			goto err;
		k = prematch(rockid, buf, ROCKIDLEN, m+rv);
		assert(k <= ROCKIDLEN);
		if (m+rv == ROCKIDLEN && k == 0) {
			/* match! */
			break;
		}
		if (k > 0) {
			/* move unmatched prefix to spill */
			if (room(spill) < k
			    && !(spill = double_ring(spill)))
				goto err;
			rs_push_ring(spill, buf, k);
		}
		m = (m+rv)-k;
		memmove(buf, &buf[k], m);
	}
	if (rs_ring_nbytes(spill) > 0)
		rs->edpspill = spill;
	else
		rs_free_ring(spill);
	return 0;
err:
	if (spill)
		rs_free_ring(spill);
	return -1;
}

/* returns:
      -1 error - no connection
       0 peer is not a rock
       1 peer is a rock
*/
int
rs_iop_probe(rs_t rs)
{
	socket_t td;
	int isrock;

	isrock = client_probe(rs);  /* closes rs->sd */
#if 1
	if (isrock)
		rs_log("edp: rock <%p> peer is a rock", rs);
	else
		rs_log("edp: rock <%p> peer is not a rock", rs);
#endif
	td = socket(AF_INET, SOCK_STREAM, 0);
	if (0 > td) {
		rserrno = errno;
		return -1;
	}
	if (rs_rebind_rock(rs, td) == INVALID_SOCKET) {
		rserrno = errno;
		return -1;
	}

	return isrock;
}

int
rs_iop_complete(rs_t rs)
{
	assert(rs->role == RS_ROLE_CLIENT
	       || rs->role == RS_ROLE_SERVER);
	if (0 > rs_xwrite(rs->sd, rockid, ROCKIDLEN))
		return -1;
	if (rs->role == RS_ROLE_CLIENT
	    && 0 > scan_for_rockid(rs))
		return -1;
	return 1;
}

int
rs_iopsrv(rs_t rs, char *z, int len, edp_result_t *result)
{
	int l;
	int rv;
	char buf[ROCKIDLEN];

	/* any which way, we're not coming back here */
	rs->state = RS_NOTCONNECTED;
	*result = EDP_NOTROCK;

	l = MIN(len, ROCKIDLEN);
	rv = recv(rs->sd, buf, l, 0);
	if (0 > rv)
		return -1;
	if (0 == rv) {
		*result = EDP_PROBE;
		if (0 > rs_xwrite(rs->sd, rockid, ROCKIDLEN))
			return -1;
		rs->state = RS_EDP; /* FIXME: good idea? */
		return 0;
	}

	if (!strncmp(rockid, buf, rv)) {
		if (rv != ROCKIDLEN) {
			/* get the rest */ 
			rv = rs_xread(rs->sd, &buf[rv], ROCKIDLEN-rv, 0);
			if (0 >= rv)
				return -1;
			if (strncmp(rockid, buf, ROCKIDLEN))
				/* FIXME: we should tolerate this */
				return -1;
		}
		*result = EDP_ROCK;
		if (0 > rs_init_connection(rs))
			return -1;
		return 0;
	}
	rs_fallback(rs);
	memcpy(z, buf, rv);
	return rv;
}
