/* 
 *  rocks/reconnect_win.c
 *
 *  Reconnection mechanism for windows rocks.
 *
 *  Copyright (C) 2001 Victor Zandy, Greg Quinn
 *  See COPYING for distribution terms.
 */

/* TODO: haven't implemented the timeout yet */

#include "rs.h"
#include "log.h"
#include "flight.h"
#include "rslist.h"		/* for rebind */
#include "rs_win.h"

#include <process.h>
#include <assert.h>

void debug_udp(rs_t rs);

/* c = a+b */
static void
tv_add(const struct timeval *a,
       const struct timeval *b,
       struct timeval *c)
{
	c->tv_sec = a->tv_sec + b->tv_sec;
	c->tv_usec = a->tv_usec + b->tv_usec;
	if (c->tv_usec >= 1000000) {
		c->tv_sec += 1;
		c->tv_usec -= 1000000;
	}
}

static void
set_timeout(rs_t rs)
{
	struct timeval tv;
	int rv;
	rv = pi_gettimeofday(&tv, NULL);
	assert(!rv);
	tv_add(&tv, &rs->lim, &rs->tout);
	rs_log("reconnect timeout at %lu sec (cur %lu)",
	       rs->tout.tv_sec, tv.tv_sec);
}

int
rs_addr_exchange(rs_t rs)
{
	int len;

	/* bind up a fresh socket in any case */
	if (rs->rec_fd != INVALID_SOCKET) {
		pi_closesocket(rs->rec_fd);
		rs->rec_fd = INVALID_SOCKET;
	}
	
	/* bind a server socket with IP address of the rock */
	rs->rec_fd = socket(AF_INET, SOCK_STREAM, 0);
	if (0 > rs->rec_fd)
		goto err;
	/* Non blocking avoids the timing problem described in
	   Stevens Network Programming V1 (2nd ed.), sec. 15.6 */
	if (0 > rs_nonblock(rs->rec_fd, 1))
		assert(0);
	if (0 > rs_reuseaddr(rs->rec_fd))
		assert(0);

	if (rs_opt_localhost) {
		/* always attempt to reconnect to the localhost
		   (this is process migration hack) */
		memset(&rs->sa_rl, 0, sizeof(rs->sa_rl));
		rs->sa_rl.sin_addr.s_addr = 0x0100007f;
	} else {
		len = sizeof(rs->sa_rl);
		if (0 > getsockname(rs->sd, (struct sockaddr *) &rs->sa_rl, &len))
			goto err;
	}

	rs->sa_rl.sin_port = htons(0);
	if (0 > bind(rs->rec_fd, (struct sockaddr *) &rs->sa_rl,
		     sizeof(rs->sa_rl)))
		goto err;
	len = sizeof(rs->sa_rl);
	if (0 > getsockname(rs->rec_fd, (struct sockaddr *) &rs->sa_rl, &len))
		goto err;
	/* we don't listen until reconnection */

	/* Exchange reconnection addresses with peer - network order */
	if (0 >= rs_xwrite(rs->sd, (iobuff_t)&rs->sa_rl, sizeof(rs->sa_rl)))
		goto err;
	if (0 >= rs_xread(rs->sd, (iobuff_t)&rs->sa_rp, sizeof(rs->sa_rp), 0))
		goto err;
	rs_log("peer reconn address: %s", inet_ntoa(rs->sa_rp.sin_addr));
	return 0;
err:
	rs->rec_fd = -1;	
	return -1;
}

void rs_wait_reconnect(rs_t rs) {

	if (WaitForSingleObject(rs->plat_dep.rec_event, INFINITE) == WAIT_FAILED)
		assert(0);
}

/* establishes the reconnection */
static socket_t reconnect(rs_t rs) {

	socket_t fd;
	int namelen;

	/* create socket pair for reconnection notification */
	memset(&rs->plat_dep.rec_notify_addr, 0, sizeof(struct sockaddr_in));
	rs->plat_dep.rec_notify_addr.sin_family = AF_INET;
	rs->plat_dep.rec_notify_addr.sin_addr.S_un.S_addr
		= htonl(INADDR_LOOPBACK);
	rs->plat_dep.rec_notify_addr.sin_port = htons(0);
	assert((rs->plat_dep.rec_notify_sd = socket(AF_INET, SOCK_DGRAM, 0))
		!= SOCKET_ERROR);
	assert((rs->plat_dep.rec_select_sd = socket(AF_INET, SOCK_DGRAM, 0))
		!= SOCKET_ERROR);
	assert(bind(rs->plat_dep.rec_select_sd,
		      (struct sockaddr *)&rs->plat_dep.rec_notify_addr,
		      sizeof(struct sockaddr_in)) != SOCKET_ERROR);
	namelen = sizeof(rs->plat_dep.rec_notify_addr);
	assert(getsockname(rs->plat_dep.rec_select_sd,
		(struct sockaddr *)&rs->plat_dep.rec_notify_addr,
		&namelen) != SOCKET_ERROR);

	/* establish connection with peer */
	fd = rs_1of2(&rs->sa_rl, &rs->sa_rp, rs->rec_fd,
	             &rs->tout, rs->role);

	return fd;
}

/* notify rocks lib that connection is back */
static void complete(rs_t rs, socket_t sd) {

	int len, rv;

	/* bind the rock to the new socket */
	rs->sd = sd;

	/* go blocking for the initialization stuff */
	/* assert(rs_nonblock(rs->sd, 0) == 0); */

    /* Heartbeat must be established and beating before in-flight
       recovery so that full fault detection is available during
       in flight recovery. */
	if (rs_opt_hb)
		if (0 > rs_hb_establish(rs->sd, rs->hb, rs->role)) {
			rs_tty_print("hb establish");
			goto failed;
		}

	/* exchange reconnection addresses */
	if (0 > rs_addr_exchange(rs)) {
		rs_tty_print("rs_addr_ex");
		goto failed;
	}

	/* update address fields */
	len = sizeof(rs->sa_locl);
	if (0 > getsockname(rs->sd, (struct sockaddr *)&rs->sa_locl, &len)) {
		rs_tty_print("getsockname");
		goto failed;
	}
	len = sizeof(rs->sa_peer);
	if (0 > getpeername(rs->sd, (struct sockaddr *)&rs->sa_peer, &len)) {
		rs_tty_print("getperrname");
		goto failed;
	}

	/* reset on close for failure detection */
	rs_reset_on_close(rs->sd, 1);

	/* finally, recover inflight data */
	if (rs_opt_flight) {
		rv = rs_inflight_recover(rs->sd, rs->ring, rs->rcvseq, rs->sndseq,
		                         &rs->maxsnd, &rs->maxrcv);
		if (0 > rv) {
			rs_tty_print("inflight_recover");
			goto failed;
		}
	}

	/* set estanblished */
	rs->state = RS_ESTABLISHED;

	/* restore event binding */
	event_rebind(rs);

	/* update the event objects */
	if (!ResetEvent(rs->plat_dep.fail_event))
		assert(0);
	if (!SetEvent(rs->plat_dep.rec_event))
		assert(0);

	/* send reconnection notification */
	assert(sendto(rs->plat_dep.rec_notify_sd, (char *)&rs, sizeof(rs), 0,
	              (struct sockaddr *)&rs->plat_dep.rec_notify_addr,
				  sizeof(struct sockaddr_in)) == sizeof(rs));

	closesocket(rs->plat_dep.rec_notify_sd);
	closesocket(rs->plat_dep.rec_select_sd);

	rs_tty_print("reconnected %d", rs->plat_dep.name);

	return;

failed:
	rs_log("reconnection completion failed");
	do_reconnect(rs);
}

void
relisten(rs_t rs)
{
	socket_t sd = INVALID_SOCKET;
	struct sockaddr_in addr;

	/* try to restart a listening sockets in one
	   shot in the same process.  bound to break
	   someday; in that case, adapt the reconnect
	   daemon to listening. */
	sd = socket(AF_INET, SOCK_STREAM, 0);
	if (sd == INVALID_SOCKET)
		goto fail;
	if (sd != rs->sd) {
		rs_rebind_rock(rs, sd);
	}
	memset(&addr, 0, sizeof(addr));
	memcpy(&addr, &rs->sa_locl, sizeof(addr));
	if (bind(rs->sd, (struct sockaddr *)&addr, sizeof(addr)) == SOCKET_ERROR)
		goto fail;
	if (listen(rs->sd, rs->backlog) == SOCKET_ERROR)
		goto fail;
	rs_log("<%d:%p> relisten ok", rs->sd, rs);
	return;
fail:
	if (sd >= 0)
		pi_closesocket(sd);
	rs_log("<%d:%p> relisten failed", rs->sd, rs);
}

/* called when connection fails */
void rs_reconnect(rs_t rs, block_t block) {

	if (rs->role == RS_ROLE_LISTEN) {
		relisten(rs);
		return;
	}

	if (!ResetEvent(rs->plat_dep.rec_event))
		assert(0);
	if (!SetEvent(rs->plat_dep.fail_event))
		assert(0);

	rs_tty_print("suspended %d", rs->plat_dep.name);

	if (block == RS_BLOCK)
		rs_wait_reconnect(rs);
}

/* this is called by control thread when reconnection is needed */
void do_reconnect(rs_t rs) {

	socket_t sd;

	assert(rs);
	rs->state = RS_SUSPENDED;
	pi_closesocket(rs->sd);
	if (rs->cb && rs->cb->suspend)
		rs->cb->suspend(rs->sd);
	set_timeout(rs);

	sd = reconnect(rs);

	/* FIXME: add handling for INVALID_SOCKET ( == reconn timeout) */
	assert(sd != INVALID_SOCKET);

	complete(rs, sd);
}