/* 
 *  rocks/rs.c
 *
 *  Sockets API implementation, except select, sockopt, and I/O.
 *
 *  Copyright (C) 2001 Victor Zandy
 *  See COPYING for distribution terms.
 */

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <assert.h>
#include <string.h>

#include "rs.h"
#include "flight.h"
#include "ring.h"
#include "log.h"
#include "rslist.h"

char *rs_roles[] = { "server", "client", "listening", "undefined" };

int rs_pid;

static void
env_options()
{
	if (getenv("RS_NOAUTH"))
		rs_opt_auth = 0;
	if (getenv("RS_NOLOG"))
		rs_opt_log = 0;
	if (getenv("RS_NOINTEROP"))
		rs_opt_interop = 0;
	if (getenv("RS_NOHB"))
		rs_opt_hb = 0;
	if (getenv("RS_NOFLIGHT"))
		rs_opt_flight = 0;
	rs_opt_ckptpath = getenv("RS_CKPTPATH");
	if (getenv("RS_LOCALHOST"))
		rs_opt_localhost = 1;
}

/* FIXME: need a beter way to iterate rocks list */
void
close_shop(void)
{
	rs_t rs;
	int i;
	for (i = 0; i < RS_MAXFD; i++)
		if ((rs = rs_lookup(i)) && rs->state == RS_ESTABLISHED)
			rs_close(rs->sd);
	rs_log("*** exit ***");
}

void
rs_init_log()
{
	char buf[32];
	if (!rs_opt_log)
		rs_startlog(NULL, RS_LOGNOLOG);
	else {
#if 1
#if 0
		/* FIXME: fork should open a new pid-based log */
		snprintf(buf, sizeof(buf), DEFAULT_LOG ".%d", getpid());
#else
		pi_snprintf(buf, sizeof(buf), DEFAULT_LOG);
#endif
		rs_startlog(buf, 0);
#else

		/* FIXME: Logging to stderr (or stdout)
		   breaks rockd, if not other programs.
		   I think the problem is the std* fd
		   becomes the connection fd.  Once the
		   connection failed, the std* fds are
		   invalid, and may get reused.  In the
		   rockd case, the fd is reused in the
		   reconnection daemon socketpair, and
		   subsequent logging messages are
		   confused for reconnection event
		   messages.*/
		assert(0);
  	        rs_startlog(NULL, 0); /* log to stderr */
#endif
	}
}

rs_t
rs_new_rs(socket_t sd, int state, int type)
{
	rs_t rs;

	rserrno = 0;
	/* assert(!rocks[sd]); Not in use */

	rs = (rs_t) malloc(sizeof(struct rs_));
	if (!rs) {
		rserrno = ENOMEM;
		return NULL;
	}
	rs->sd = sd;
	rs->doedp = rs_opt_interop;
	rs->nonblock = 0;
	rs->rec_fd = -1;
	rs->rec_pid = 0;
	rs->type = type;
	rs->state = state;
	rs->ring = NULL;
	rs->rcvseq = 0;
	rs->sndseq = 0;
	rs->maxrcv = 0;
	rs->maxsnd = 0;
	rs->role = RS_ROLE_UNDEF;
	rs->hb = NULL;
	rs->edpspill = NULL;
	rs->clospill = NULL;
	rs->booger = INVALID_SOCKET;
	rs->shmid = 0;
	rs->shm = NULL;
	rs->lim.tv_sec = 3*24*60*60; /* 3 days */
	rs->lim.tv_usec = 0;
	rs->tout.tv_sec = rs->tout.tv_usec = 0;
	rs->refcnt = 1;
	rs->cb = NULL;
	rs_platform_socket(rs);
	rs_log("added %d to rocks", sd);
	return rs;
}

static void
free_rs(rs_t rs)
{
	if (!rs)
		return;
	rs_platform_free(rs);
	if (rs->ring)
		rs_free_ring(rs->ring);
	if (rs->hb)
		rs_free_heartbeat(rs->hb);
	if (rs->edpspill)
		rs_free_ring(rs->edpspill);
	if (rs->clospill)
		rs_free_ring(rs->clospill);
	if (rs->rec_fd)
		pi_closesocket(rs->rec_fd);
	free(rs);
}

void
rs_init()
{
	rs_pid = pi_getpid();
	env_options();
	if (0 > rs_init_sys()) {
		fprintf(stderr, "Can't initialize reliable sockets\n");
		exit(1);
	}
	rs_init_signal();
	rs_platform_init();
	rs_init_log();
	rs_log("*** Rocks loaded ***");

	if (rs_in_exec()) {
		/* FIXME: Necessary mode switch? */
		rs_mode_native();		
		rs_restore_exec();
		rs_mode_pop();
	}

	if (rs_opt_hb) {
		if (0 > rs_init_heartbeat()) {
			fprintf(stderr, "Can't initialize reliable sockets\n");
			exit(1);
		}
	}	  
	if (0 > atexit(close_shop)) {
		fprintf(stderr, "Can't initialize reliable sockets exit\n");
		exit(1);
	}
#ifndef NO_AUTH
	if (rs_opt_auth && 0 > rs_init_crypt()) {
		fprintf(stderr, "Can't initialize reliable sockets\n");
		exit(1);
	}
#endif
	rs_mode_native();
	rs_check_open_fds();
	rs_mode_pop();
}

/* Sets in FDSET the descriptors for all established or suspended
   reliable sockets.  Returns the highest descriptor set plus 1, or 0
   if none are established. */
int
rs_fdset(fd_set *fdset)
{
	int d, max;
	rs_t rs;

	max = 0;
	for (d = 0; d < RS_MAXFD; d++)
		if ((rs = rs_lookup(d))
		    && rs->state != RS_NOTCONNECTED
		    && rs->state != RS_EDP) {
			/* FIXME: This is booger (UDP) code.  It keeps the
			   heartbeat handler from thinking this rock
			   has a heartbeat. */
			if (RS_ROLE_LISTEN == rs->role)
				continue;
			FD_SET(d, fdset);
			max = d+1;
		}
	return max;
}

void
rs_fallback(rs_t rs)
{
	rs_log("<%d:%p> fallback to ordinary socket", rs->sd, rs);
	rs_unrock(rs);
	free_rs(rs);
}

socket_t
rs_accept(socket_t srv_sd, struct sockaddr *iaddr, int *addrlen)
{
	salen_t len;
	rs_t srv_rs, rs;
	socket_t sd;
	struct sockaddr_in addr;

	rs_log("In rs_accept");
	rserrno = 0;
	srv_rs = rs_lookup(srv_sd);
	if (!srv_rs) {
		rserrno = EINVAL;
		rs_log("rs_accept error at lookup");
		return -1;
	}
rs_accept_retry:
	len = sizeof(addr);
	sd = accept(srv_rs->sd, (struct sockaddr *)&addr, &len);
	if (0 > sd) {
		if (errno == EINTR)
			goto rs_accept_retry;
		rserrno = errno;
		rs_log("rs_accept error at accept %s", strerror(errno));
		return INVALID_SOCKET;
	}

	rs = rs_new_rs(sd, RS_NOTCONNECTED, SOCK_STREAM);
	if (!rs)
		return INVALID_SOCKET;
	rs->role = RS_ROLE_SERVER;
	if (0 > rs_reuseaddr(rs->sd)) {
		rserrno = errno;
		return INVALID_SOCKET;
	}
	memcpy(&rs->sa_locl, &srv_rs->sa_locl, sizeof(rs->sa_locl));
	memcpy(&rs->sa_peer, &addr, sizeof(rs->sa_peer));
	if (addrlen)
		*addrlen = MIN(*addrlen, sizeof(addr));
	if (iaddr && addrlen)
		memcpy(iaddr, &addr, *addrlen);

	rs_log("<%d:%p> accept -> <%d:%p>", srv_rs->sd, srv_rs, rs->sd, rs);
	/* don't init until we see some client proof of rockhood */
	rs->state = RS_EDP;
	return rs_add_rock(rs);
}

int
rs_close(socket_t sd)
{
	rs_t rs;
	int cnt;

	rserrno = 0;

	rs_log("rs_close %d", sd);
	if (sd == rs_logfileno()) {
		rs_log("ignoring application attempt to shut me up!");
		return 0;
	}

	rs = rs_lookup(sd);
	if (!rs) {
		rserrno = EINVAL;
		return -1;
	}

	/* duplicated descriptor */
	if (rs->shm) {
		rs_log("locking shm");
		rs_shm_lock(rs->shm);
		rs_log("shm locked");
		cnt = --rs->shm->refcnt;
		rs_shm_unlock(rs->shm);
	} else
		cnt = --rs->refcnt;
	if (cnt >= 1) {
		/* local close - don't kill the connection */
		rs_remove_node(sd);
		pi_closesocket(sd);
		return 0;
	}

	if (rs->booger != INVALID_SOCKET) {
		/* FIXME: Booger code; doesn't close listeners. */
		socket_t b = rs->booger;
		rs->booger = INVALID_SOCKET;
		return rs_close(b);
	}
        rs->state = RS_NOTCONNECTED;
	if (rs->hb)
		rs_hb_cancel(rs->hb);
	rs_remove_node(sd);
	if (rs_rock_is_shared(rs))
		rs_shm_detach(rs);
	rs_reset_on_close(rs->sd, 0);
	pi_closesocket(rs->sd);
	rs_log("<%d:%p> closed", sd, rs);
	free_rs(rs);
	return 0;
}

int
rs_shutdown(socket_t sd, int how)
{
	if (how == 2)
		return rs_close(sd);
	assert(0);
	return 0;
}

int
rs_init_connection(rs_t rs)
{
	salen_t len;

	/* go blocking for the initialization stuff */
	assert(rs_nonblock(rs->sd, 0) == 0);

	rs_iop_complete(rs);

	len = sizeof(rs->sa_locl);
	if (0 > getsockname(rs->sd, (struct sockaddr *) &rs->sa_locl, &len)) {
		rserrno = errno;
		return SOCKET_ERROR;
	}

	rs_log("<%d:%p> locl recon is %s", rs->sd, rs, rs_ipstr(&rs->sa_locl));
	rs_log("<%d:%p> peer recon is %s", rs->sd, rs, rs_ipstr(&rs->sa_peer));
	if (0 > rs_reuseaddr(rs->sd)) {
		rserrno = errno;
		return -1;
	}
	/* NO_DELAY reduces connect time. */
	if (0 > rs_nodelay(rs->sd, 1)) {
		rserrno = errno;
		return -1;
	}
#ifndef NO_AUTH
	/* Exchange session key */
	rs_log("<%d:%p> init pre auth", rs->sd, rs);
	if (rs_opt_auth) {
		rs->key = rs_key_exchange(rs->sd);
		if (!rs->key) {
			rs_log("Unable to establish key");
			rserrno = ERSINIT;
			return -1;
		}
	}
#endif
	/* exchange reconnection addresses */
	if (0 > rs_addr_exchange(rs)) {
		rs_log("Unable to exchange reconnection address");
		rserrno = ERSINIT;
		return -1;
	}

	if (rs_opt_flight) {
		if (0 > rs_inflight_limits(rs->sd, &rs->maxsnd, &rs->maxrcv)) {
			rserrno = ERSINIT;
			return -1;
		}
		rs->ring = rs_new_ring(rs->maxsnd);
		if (!rs->ring) {
			rserrno = ENOMEM;
			return -1;
		}
	} else {
		/* FIXME: Do something better than this. */
		/* See usage of maxsnd in rw.c */
		rs->maxsnd = 1000000;
	}
	if (0 > rs_reset_on_close(rs->sd, 1)) {
		rserrno = ERSINIT;
		return -1;
	}
	if (rs_opt_hb) {
		rs->hb = rs_new_heartbeat(rs);
		if (!rs->hb) {
			rserrno = ENOMEM;
			free(rs);
			return -1;
		}
		if (0 > rs_hb_establish(rs->sd, rs->hb, rs->role)) {
			rserrno = ERSINIT;
			return -1;
		}
	}
	/* FIXME: Restore application NO_DELAY state. */
	if (0 > rs_nodelay(rs->sd, 1)) {
		rserrno = errno;
		return -1;
	}
	rs->state = RS_ESTABLISHED;
	assert(rs_platform_connect(rs) == 0);
	return 0;
}

socket_t
rs_socket(int domain, int type, int protocol)
{
	socket_t sd;
	rs_t rs;

	rserrno = 0;

	if (!rs_opt_udp && SOCK_DGRAM == type) {
		if ((sd = socket(domain, type, protocol)) < 0)
			return INVALID_SOCKET;
		else
			return rs_add_nonrock(sd);
	}

	/* Always create a SOCK_STREAM, but store the requested type
	   in the rs so we can give it the behavior later. */
	sd = socket(domain, SOCK_STREAM, protocol);
	if (0 > sd) {
		rserrno = errno;
		return INVALID_SOCKET;
	}
	if (domain != AF_INET)
		/* Default to ordinary socket */
		return rs_add_nonrock(sd);
	if (0 > rs_reuseaddr(sd)) {
		rserrno = errno;
		return INVALID_SOCKET;
	}
	rs = rs_new_rs(sd, RS_NOTCONNECTED, type);
	if (!rs)
		return INVALID_SOCKET;

	rs_log("<%d:%p> new rock", sd, rs);
	return rs_add_rock(rs);
}

int
rs_bind(socket_t sd, const struct sockaddr *iaddr, salen_t addrlen)
{
	struct sockaddr_in *addr;
	rs_t rs;
	int len;

	rserrno = 0;
	if (iaddr->sa_family != AF_INET) {
		rserrno = EPROTONOSUPPORT;
		return -1;
	}
	addr = (struct sockaddr_in *) iaddr;
	rs = rs_lookup(sd);
	if (!rs) {
		rserrno = EINVAL;
		return -1;
	}
	if (rs->state != RS_NOTCONNECTED) {
		rserrno = EISCONN;
		return -1;
	}
	if (0 > bind(rs->sd, iaddr, addrlen)) {
		rserrno = errno;
		return -1;
	}
	len = sizeof(rs->sa_locl);
	if (0 > getsockname(rs->sd, (struct sockaddr *) &rs->sa_locl, &len)) {
		rserrno = errno;
		return -1;
	}
	return 0;
}

int
rs_listen(socket_t sd, int backlog)
{
	rs_t rs;

	rserrno = 0;
	rs = rs_lookup(sd);
	if (!rs) {
		rserrno = EINVAL;
		return SOCKET_ERROR;
	}
	if (rs->state != RS_NOTCONNECTED) {
		rserrno = EISCONN;
		return SOCKET_ERROR;
	}
	if (0 > listen(rs->sd, backlog)) {
		rserrno = errno;
		return SOCKET_ERROR;
	}
	if (0 > rs_reuseaddr(rs->sd)) {
		rserrno = errno;
		return SOCKET_ERROR;
	}
	rs->role = RS_ROLE_LISTEN;
	rs->backlog = backlog;
	rs_log("<%d:%p> listening on %d", rs->sd, rs,
	       ntohs(rs->sa_locl.sin_port));
	return 0;
}

int
rs_connect(socket_t sd, const struct sockaddr *iaddr, salen_t peerlen)
{
	struct sockaddr_in *peer;
	rs_t rs;

	rserrno = 0;
	rs = rs_lookup(sd);
	if (!rs) {
		rserrno = EINVAL;
		return -1;
	}
	if (iaddr->sa_family != AF_INET) {
		rserrno = EPROTONOSUPPORT;
		return -1;
	}
	peer = (struct sockaddr_in *) iaddr;
	if (rs->state != RS_NOTCONNECTED) {
		rserrno = EISCONN;
		return -1;
	}

	/* write peer address to address structure */
	memcpy(&rs->sa_peer, peer, sizeof(rs->sa_peer));

	rs_log("<%d:%p> connecting to %s", rs->sd, rs, rs_ipstr(peer));
	if (0 > connect(rs->sd, (struct sockaddr *)peer, peerlen))
	{
		rserrno = pi_socket_errno();	
		return SOCKET_ERROR;
	}

	return rs_continue_connect(rs);
}

int
rs_continue_connect(rs_t rs)
{
	int isrock;

	rs->role = RS_ROLE_CLIENT;

	if (!rs->doedp) {
		rs_log("<%d:%p> connect to presumed rock", rs->sd, rs);
		return rs_init_connection(rs);
	}

	rs_log("start client probe for <%d:%p>", rs->sd, rs);
	isrock = rs_iop_probe(rs);
	if (0 > isrock)
		return SOCKET_ERROR;
	rs_log("<%d:%p> connect to %s", rs->sd, rs,
	       isrock ? "rock" : "non-rock");
	if (!isrock) {
          int x;
		socket_t sock = rs->sd;
		struct sockaddr_in addr;
		memcpy(&addr, &rs->sa_peer, sizeof(rs->sa_peer));
		rs_fallback(rs);
		x = connect(sock, (struct sockaddr *)&addr,
			       sizeof(addr));
          rserrno = pi_socket_errno();
          return x;
	} else {
		int x = connect(rs->sd, (struct sockaddr *)&rs->sa_peer,
				sizeof(rs->sa_peer));
		if (x == SOCKET_ERROR)
			return x;
		else
			return rs_init_connection(rs);
	}
}

int
rs_getsockname(socket_t sd, struct sockaddr *name, salen_t *namelen)
{
	rs_t rs;

	rserrno = 0;
	rs = rs_lookup(sd);
	if (!rs) {
		rserrno = EINVAL;
		rs_log("Failed to getsockname");
		return -1;
	}
	memcpy(name, &rs->sa_locl, sizeof(rs->sa_locl));
	*namelen = sizeof(rs->sa_locl);
	return 0;
}

int
rs_getpeername(socket_t sd, struct sockaddr *name, salen_t *namelen)
{
	rs_t rs;

	rserrno = 0;
	rs = rs_lookup(sd);
	if (!rs) {
		rserrno = EINVAL;
		rs_log("Failed to getsockname");
		return -1;
	}

	rserrno = 0;
	memcpy(name, &rs->sa_peer, sizeof(rs->sa_peer));
	*namelen = sizeof(rs->sa_peer);
	return 0;
}

static void
exit_cleanup()
{
	rs_t rs;
	int fd;
	for (fd = 0; fd < RS_MAXFD; fd++) {
		rs = rs_lookup(fd);
		if (!rs)
			continue;
		/*
		if (rs->state == RS_SUSPENDED
		    && !rs_rock_is_shared(rs))
			kill(rs->rec_pid, SIGKILL);
		*/
	}
}

void
rs_exit(int status)
{
	exit_cleanup();
	exit(status);
}
