/* 
 *  rocks/rs.h
 *
 *  Copyright (C) 2001 Victor Zandy
 *  See COPYING for distribution terms.
 */

#ifndef _RS_H_
#define _RS_H_

#include "pi.h"
#include "ring.h"

extern int rserrno;
const char *rserr();
extern int rs_pid;   /* process id we think we are in; updated across forks */

#define MIN(x,y) ((x) < (y) ? (x) : (y))
#define MAX(x,y) ((x) > (y) ? (x) : (y))

#define RS_MAXFD 1024 /* FIXME: Use the system limit */

/* Options */
extern int rs_opt_auth;
extern int rs_opt_interop;
extern int rs_opt_hb;
extern int rs_opt_log;
extern int rs_opt_flight;
extern int rs_opt_udp;
extern int rs_opt_alarm_period;
extern int rs_opt_max_alarm_misses;
extern int rs_opt_auth_timeout;
extern int rs_opt_rec_max_timeouts;
extern int rs_opt_localhost;
extern char *rs_opt_ckptpath;

#define ERSUNUSED       256    /* unused */
#define ERSINIT         257    /* rs socket init failed */

typedef struct rs_ * rs_t;

int rs_fork();
int rs_vfork();
void rs_exit(int status);
socket_t rs_socket(int domain, int type, int protocol);
int rs_bind(socket_t sd, const struct sockaddr *iaddr, salen_t addrlen);
int rs_listen(socket_t sd, int backlog);
socket_t rs_accept(socket_t srv_sd, struct sockaddr *addr, int *addrlen);
int rs_connect(socket_t sd, const struct sockaddr *iaddr, salen_t addrlen);
int rs_continue_connect(rs_t);
int rs_complete_connect(rs_t rs);
int rs_close(socket_t fd);
int rs_shutdown(socket_t sd, int how);
int rs_fdset(fd_set *);
int rs_getsockname(socket_t s, struct sockaddr *name, salen_t *namelen);
int rs_getpeername(socket_t s, struct sockaddr *name, salen_t *namelen);
ioret_t rs_recv(socket_t sd, iobuff_t buf, iosize_t len, int flags);
ioret_t rs_send(socket_t sd, iobuff_t buf, iosize_t len, int flags);
ioret_t rs_recvfrom(socket_t sd, iobuff_t buf, iosize_t len, int flags,
		    struct sockaddr *from, salen_t *fromlen);
ioret_t rs_sendto(socket_t sd, const iobuff_t msg, iosize_t len, int flags,
	          const struct sockaddr *to, salen_t tolen);
int rs_select(int n, fd_set *rs, fd_set *ws, fd_set *es,
	      struct timeval *tv);

/* Overloaded syscalls needed for our sanity, not for the sockets API */

/* util.c */
typedef struct rocklist *rocklist_t;
struct rocklist {
	rs_t rs;
	rocklist_t next;
};
int rs_xwrite(socket_t sd, iobuff_t buf, iosize_t len);
int rs_xread(socket_t sd, iobuff_t buf, iosize_t len, unsigned long ms);
int rs_reset_on_close(socket_t sd, int onoff);
int rs_recover_bad_rocks(int n, fd_set *fds);
int rs_reuseaddr(socket_t sd);
int rs_nodelay(socket_t sd, int onoff);
int rs_settcpbuf(socket_t sd, int type, int size);
int rs_nonblock(socket_t sd, int on);
int rs_waitread(socket_t sd, unsigned long ms);
void rs_tv_diff(const struct timeval *a,
		const struct timeval *b,
		struct timeval *c);
void rs_logbytes(char *bytes, int len);
int rs_rocklist_insert(rocklist_t *head, rs_t rs);
int rs_rocklist_remove(rocklist_t *head, rs_t rs);
rs_t rs_rocklist_findsa(rocklist_t head, struct sockaddr_in *sa);
void rs_rocklist_free(rocklist_t *head);

int rs_init_sys();
void rs_init_signal();
void rs_init();
int rs_setsockopt(socket_t s, int level, int optname,
		  iobuff_t optval, salen_t optlen);
int rs_ioctl(socket_t fd, ioctlcmd_t cmd, ioctlarg_t arg);

/* Internal library functions below? */

#ifndef NO_AUTH
/* Crypto implementation dependent declarations */
typedef struct rs_key *rs_key_t;
int rs_init_crypt();
void rs_key_free(rs_key_t key);
rs_key_t rs_key_exchange(int sock);
int rs_mutual_auth(rs_key_t key, int sock);
int rs_key_save(rs_key_t key, int fd);
rs_key_t rs_key_restore(int fd);

/* Common crypto functions */  
int rs_authenticate(rs_key_t key, int sock);
#else
typedef void *rs_key_t;
#endif /* NO_AUTH */

typedef struct hb * hb_t;
typedef enum { EDP_NOTROCK, EDP_PROBE, EDP_ROCK } edp_result_t;

typedef struct recstate recstate_t;
struct recstate {
	socket_t sda, sdp;        /* active and passive socks */
	socket_t sdn;             /* new incoming connection on sda */
	socket_t sdc;             /* final connection */
	int sent;                 /* Have we (if server) sent the byte? */
	int est;                  /* Established connections bitset */
	int timeouts;    	  /* Timeout counter  */
};

typedef enum { RS_NOTCONNECTED,
	       RS_ESTABLISHED,
	       RS_EDP,
	       RS_SUSPENDED,
	       RS_HUNGUP } rs_state;
typedef enum { RS_ROLE_SERVER, RS_ROLE_CLIENT, RS_ROLE_LISTEN, RS_ROLE_UNDEF } rs_role;
extern char *rs_roles[]; /* If you add a role, update rs_roles. */

typedef struct shm *shm_t;
typedef enum { RS_BLOCK, RS_NOBLOCK } block_t;

struct callbacks {
	void (*suspend)(socket_t rock);
};

struct rs_ {
	rs_state state;
	int refcnt;		    /* reference count */
	int type;                   /* socket type */
	int doedp;		    /* enhancement detection upon connect? */
     int nonblock;				/* is socket nonblocking */
	ring_t ring;                /* in flight buffer */
	unsigned rcvseq;            /* seq # of next byte to be passed to app */
	unsigned sndseq;            /* seq # of next byte to be passed to kernel */
	unsigned maxrcv;            /* TCP socket recv buffer size */
	unsigned maxsnd;            /* TCP socket send buffer size */
	socket_t sd;                /* socket descriptor */
	struct sockaddr_in sa_locl; /* current address of this endpoint */
	struct sockaddr_in sa_peer; /* current address of peer endpoint */
	struct sockaddr_in sa_rl;   /* local reconnection address */
	struct sockaddr_in sa_rp;   /* peer reconnection address */
	int rec_pid;		    /* reconnection process */
	socket_t rec_fd;            /* socket to reconnection process */
	hb_t hb;                    /* heartbeat state */
	rs_key_t key;               /* authentication key */
	recstate_t rec;             /* reconnection state */
	struct timeval lim;         /* reconnection time limit */
	struct timeval tout;        /* current reconnection timeout alarm */
	rs_role role;		    /* role in connection */
	int backlog;		    /* backlog for listening sockets */
	ring_t edpspill;	    /* edp client recv buffer */
	ring_t clospill;	    /* close catcher */
	socket_t booger;                 /* as the name implies,
									    this needs to be cleaned up */
	int shmid;                  /* if shared rock, shm identifier */
	shm_t shm;                  /* if shared rock, shared state */
	struct callbacks *cb;
	unsigned long finode;	/* socket inode; not maintained over rock life */
	rs_platform_t plat_dep;		/* platform dependent info */
};

int rs_platform_init();
int rs_platform_socket(rs_t);
int rs_platform_connect(rs_t);
int rs_platform_free(rs_t);

/* shm.c */
struct shm {
	int lfd;                /* lock file descriptor */
	int hb_owner;	        /* Process responsible for hbs */
	int hb_count;           /* Incremented each beat by owner */
	int refcnt;		/* reference count from rs_t */
};

typedef enum { RS_MODE_RS, RS_MODE_NATIVE } rs_mode_t;
void rs_mode_push(rs_mode_t m);
void rs_mode_native();
void rs_mode_pop();

char *rs_ipstr(struct sockaddr_in *addr);

rs_t rs_new_rs(socket_t, int, int);
rs_t rs_lookup(socket_t fd);
void rs_reconnect(rs_t rs, block_t block);
void rs_rec_complete(rs_t rs, block_t block);
int rs_addr_exchange(rs_t rs);
void rs_wait_reconnect(rs_t rs);

int rs_init_connection(rs_t rs);
void rs_init_log();

/* FIXME: These need to be here due to their appearance in rs.c */
int rs_in_exec();
void rs_restore_exec();
void rs_check_open_fds();

/* Heartbeat */
int rs_init_heartbeat();
hb_t rs_new_heartbeat();
void rs_free_heartbeat(hb_t hb);
int rs_hb_establish(socket_t rock, hb_t hb, rs_role role);
int rs_hb_cancel(hb_t hb);
void rs_become_hb_owner(hb_t hb);
void rs_hb_init_shm(rs_t rs);

/* Interoperability */
void rs_free_iop(rs_t rs);
int rs_iopsrv(rs_t rs, char *z, int len, edp_result_t *result);
int rs_iop_probe(rs_t);
int rs_iop_complete(rs_t);
void rs_fallback(rs_t rs);

/* Interprocess rock sharing */
int rs_rock_is_shared(rs_t rs);
int rs_shm_has_one_owner(rs_t rs);
int rs_shm_create(rs_t rs);
int rs_shm_attach(rs_t rs);
void rs_shm_detach(rs_t rs);
void rs_shm_lock(shm_t shm);
void rs_shm_unlock(shm_t shm);

/* 1 of 2 connection establishment */
socket_t rs_1of2(struct sockaddr_in *locl, struct sockaddr_in *peer, 
	    socket_t ls, struct timeval *lim, rs_role role);

#endif /* _RS_H_ */

/* sockaddr.c */
int rs_xwrite_ipaddr(socket_t fd, struct sockaddr_in *addr);
int rs_xread_ipaddr(socket_t fd, struct sockaddr_in *addr);
