/*
 * UNIX-specific rocks functions
 */

#include "rs_unix.h"
#include "log.h"
#include "ring.h"
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <assert.h>

/* do dup (NEW < 0) or dup2 (NEW >= 0) */
static int
do_dup(int old, int new)
{
	rs_t rs;
	rs = rs_lookup(old);
	if (!rs) {
		rserrno = EBADF;
		return -1;
	}
	if (new >= 0) {
		/* dup2 closes new if necessary */
		if (rs_lookup(new))
			rs_close(new);
		new = dup2(old, new);
	} else
		new = dup(old);
	if (0 > new)
		return -1;
	rs_share_rockfd(old, new);
	return new;
}

int
rs_dup(int old)
{
	return do_dup(old, -1);
}

int
rs_dup2(int old, int new)
{
	return do_dup(old, new);
}

/* this was formerly in sockopt.c */
int
rs_fcntl(int fd, int cmd, long arg)
{
	if (cmd == F_SETFL && (arg & O_NONBLOCK)) {
		rs_log("Not ignoring app NONBLOCK");
		fcntl(fd, cmd, O_NONBLOCK);
		return 0;
	} else if (cmd == F_DUPFD) {
		rs_log("SHIT! Ignoring app F_DUPFD");
		return 0;
	}
	return fcntl(fd, cmd, arg);
}

int 
rs_save(rs_t rs, int fd)
{
	rs_log("saving rock %d", rs->sd);
	if (0 > rs_xwrite(fd, rs, sizeof(*rs)))
		return -1;
	if (rs_opt_flight && 0 > rs_ring_save(rs->ring, fd))
		return -1;
	if (rs_opt_hb && 0 > rs_hb_save(rs->hb, fd))
		return -1;
#ifndef NO_AUTH
	if (rs_opt_auth && 0 > rs_key_save(rs->key, fd))
		return -1;
#endif
	return 0;
}

rs_t
rs_restore(int fd)
{
	struct rs_ r;
	rs_t rs;
	if (0 >= rs_xread(fd, &r, sizeof(r), 0))
		return NULL;
	rs = rs_new_rs(r.sd, r.state, r.type);
	*rs = r;
	if (!rs)
		return NULL;
	if (rs_rock_is_shared(rs))
		if (0 > rs_shm_attach(rs)) {
			rs_log("Error restoring exec shm");
			return NULL;
		}
	if (rs_opt_flight) {
		rs->ring = rs_ring_restore(fd);
		if (!rs->ring) {
			rs_log("Error restoring exec rock ring");
			return NULL;
		}
	}
	if (rs_opt_hb) {
		rs->hb = rs_hb_restore(rs, fd);
		if (!rs->hb) {
			rs_log("Error restoring exec rock hb");
			return NULL;
		}
	}
#ifndef NO_AUTH
	if (rs_opt_auth) {
		rs->key = rs_key_restore(fd);
		if (!rs->key) {
			rs_log("Error restoring exec rock key");
			return NULL;
		}
	}
#endif
	return rs;
}

ssize_t
rs_readv(int sd, const struct iovec *iov, int iovcnt)
{
	int i;
	int rv;
	int n;

	n = 0;
	for (i = 0; i < iovcnt; i++) {
		rv = rs_read(sd, iov[i].iov_base, iov[i].iov_len);
		if (0 > rv)
			return rv;
		n += rv;
		if (rv < iov[i].iov_len)
			break;
	}
	return n;
}

ssize_t
rs_writev(int sd, const struct iovec *iov, int iovcnt)
{
	int i;
	int rv;
	int n;

	n = 0;
	for (i = 0; i < iovcnt; i++) {
		rv = rs_write(sd, iov[i].iov_base, iov[i].iov_len);
		if (0 > rv)
			return rv;
		n += rv;
		if (rv < iov[i].iov_len)
			break;
	}
	return n;
}

int
rs_recvmsg(int sd, struct msghdr *msg, int flags)
{
	assert(0);
	return 0;
}

int
rs_sendmsg(int sd, const struct msghdr *msg, int flags)
{
	assert(0);
	return 0;
}

static int
is_tcp_inode(unsigned long inode)
{
	FILE *fp;
	char buf[1024];
	int n = 0;
	int ret = 0;
	unsigned long i;

	fp = fopen("/proc/net/tcp", "r");
	if (!fp) {
		perror("open /proc/net/tcp");
		assert(0);
	}
	while (fgets(buf, sizeof(buf), fp)) {
		if (!n++)
			continue; /* skip first line */
		sscanf(buf,
		       "%*s %*x:%*x %*x:%*x %*x %*s %*s %*s %*s %*s %lu",
		       &i);
		/* the kernel prints ports in host order;
		   see /usr/src/linux/net/ipv4/tcp_ipv4.c:^get_tcp_sock */
		if (i == inode) {
			ret = 1;
			break;
		}
	}
	fclose(fp);
	return ret;
}

static socket_t
fake_accept(socket_t sd, unsigned long inode)
{
	struct sockaddr_in addr;
	socklen_t len;
	rs_t rs;
	int i;

	rs_log("In fake accept for %d", sd);

	/* have we seen this one already? */
	for (i = 0; i < RS_MAXFD; i++)
		if (rs_lookup(i) && rs_lookup(i)->finode == inode) {
			rs_log("fake accept sharing %d with %d", sd, i);
			rs_share_rockfd(i, sd);
			return 0;
		}
		
	len = sizeof(addr);
	if (rs_lookup(sd)) {
		rs_log("already have rock for %d", sd);
		return 0; /* already got it */
	}
	rs = rs_new_rs(sd, RS_NOTCONNECTED, SOCK_STREAM);
	if (!rs)
		return -1;
	rs->role = RS_ROLE_SERVER;
	rs->finode = inode;
	len = sizeof(rs->sa_locl);
	if (0 > getsockname(sd, (struct sockaddr*)&rs->sa_locl, &len))
		return -1;
	len = sizeof(rs->sa_peer);
	if (0 > getpeername(sd, (struct sockaddr*)&rs->sa_peer, &len))
		return -1;
	if (0 > rs_reuseaddr(sd)) {
		rserrno = errno;
		return -1;
	}
	rs->state = RS_EDP;
	return sd;
}

void
rs_check_open_fds()
{
	DIR *dir;
	struct dirent *e;
	char s[1024];
	char buf[1024];
	int fd, rv;
	unsigned long inode;
	int pid;

	pid = getpid();
	snprintf(s, sizeof(s), "/proc/%d/fd", pid);
	dir = opendir(s);
	if (!dir) {
		fprintf(stderr, "cannot open %s\n", s);
		return;
	}
	while ((e = readdir(dir))) {
		snprintf(s, sizeof(s), "/proc/%d/fd/%s", pid, e->d_name);
		rv = readlink(s, buf, sizeof(buf));
		if (0 > rv)
			continue;
		if (rv >= sizeof(buf)) {
			fprintf(stderr, "buffer overflow\n");
			assert(0);
		}
		buf[rv] = '\0';
		fd = atoi(e->d_name);
		if (1 != sscanf(buf, "socket:[%lu]", &inode))
			continue;
		if (!is_tcp_inode(inode))
			continue;
		/*	Assume for now but generally incorrectly
			that open tcp fds are newly accepted connections */
		if (0 > fake_accept(fd, inode))
			rs_log("failed to fake accept %d\n", fd);
	}
	closedir(dir);
}

char *
rs_getargv0()
{
	static char *unknown = "<indeterminate command>";
	int fd, rv;
	static char buf[1024];

	fd = open("/proc/self/cmdline", O_RDONLY);
	if (0 > fd)
		return unknown;
	rv = read(fd, buf, sizeof(buf)-1);
	close(fd);
	if (0 > rv)
		return unknown;
	buf[rv] = '\0';
	return buf;
}
