/* 
 *  rocks/sys_win.c
 *  
 *  System call redirection.
 */

#include "rslist.h"
#include "log.h"
#include "rs_win.h"

#include <psapi.h>
#include <dbghelp.h>
#include <tlhelp32.h>
#include <stdlib.h>
#include <assert.h>

#define MAXPATH 256
#define NUMHOOKS 40

/* we need to keep a list of all hooked functions */
static struct hook {
	PROC old_proc;
	PROC new_proc;
	const char *old_module_name;
}
hooks[NUMHOOKS];
int num_hooks = 0;

/* modules for overridden functions */
typedef struct {
	HMODULE module;
	PSTR module_name;
} module_t;

static module_t winsock_dll = {NULL, "ws2_32.dll"};
static module_t old_winsock_dll = {NULL, "wsock32.dll"};
static module_t kernel_dll = {NULL, "kernel32.dll"};
static module_t user_dll = {NULL, "user32.dll"};

static DWORD winsock_dll_end;

void replace_rs_functions(HMODULE caller_module);

int __stdcall rs_win_WSAStartup(WORD version, LPWSADATA data) {

	int rv;

	rv = WSAStartup(version, data);

	return rv;
}

SOCKET __stdcall rs_win_socket(int af, int type, int protocol) {

	return rs_socket(af, type, protocol);
}

int __stdcall rs_win_connect(SOCKET s, const struct sockaddr *name, int namelen) {
	
	int rv, err;
	SOCKET sock = rs_lookup_socket(s);

	if (rs_lookup(s)) {
		rv = rs_connect(s, name, namelen);
		err = rserrno;
	}
	else {
		rv = connect(sock, name, namelen);
		err = WSAGetLastError();
	}
	
	WSASetLastError(err);
	if (rv == SOCKET_ERROR && WSAGetLastError() == WSAEINVAL) {
		struct sockaddr_in sa;
		int sz = sizeof(sa);
		if (getpeername(sock, (struct sockaddr *)&sa, &sz) == 0 ||
				WSAGetLastError() != WSAENOTCONN) {
			WSASetLastError(WSAEISCONN);
			rs_log("Had to fix errno value!");
		}
	}

	assert(WSAGetLastError() != WSAEINVAL);

	return rv;
}

int __stdcall rs_win_closesocket(SOCKET s) {

	if (rs_lookup(s))
		return rs_close(s);
	else
		return closesocket(rs_lookup_socket(s));
}

int __stdcall rs_win_recv(SOCKET s, char *buf, int len, int flags) {

	int rv;
	if (rs_lookup(s))
		rv = rs_recv(s, buf, len, flags);
	else
		rv = recv(rs_lookup_socket(s), buf, len, flags);

	return rv;
}

int __stdcall rs_win_send(SOCKET s, char *buf, int len, int flags) {

	int rv;

	if (rs_lookup(s))
		rv = rs_send(s, buf, len, flags);
	else
		rv = send(rs_lookup_socket(s), buf, len, flags);

	return rv;
}

SOCKET __stdcall rs_win_accept(SOCKET sd, struct sockaddr *iaddr, int *addrlen) {

	if (rs_lookup(sd))
		return rs_accept(sd, iaddr, addrlen);
	else
		return accept(sd, iaddr, addrlen);
}

int __stdcall rs_win_bind(SOCKET sd, struct sockaddr *iaddr, int addrlen) {

	if (rs_lookup(sd))
		return rs_bind(sd, iaddr, addrlen);
	else
		return bind(sd, iaddr, addrlen);
}

int __stdcall rs_win_listen(SOCKET sd, int backlog) {

	if (rs_lookup(sd))
		return rs_listen(sd, backlog);
	else
		return listen(sd, backlog);
}

int __stdcall rs_win_select(int n, fd_set *rs, fd_set *ws, fd_set *es,
                            struct timeval *tv) {

	return rs_select(n, rs, ws, es, tv);
}

int __stdcall rs_win_getsockname(SOCKET sd, struct sockaddr *name,
								  int *namelen) {

	rs_log("getsockname: name=0x%x", sd);
	if (rs_lookup(sd))
		return rs_getsockname(sd, name, namelen);
	else
		return getsockname(rs_lookup_socket(sd), name, namelen);
}

int __stdcall rs_win_getpeername(SOCKET sd, struct sockaddr *name,
								  int *namelen) {

	rs_log("getpeername: name=0x%x", sd);
	if (rs_lookup(sd))
		return rs_getpeername(sd, name, namelen);
	else
		return getpeername(rs_lookup_socket(sd), name, namelen);
}

int __stdcall rs_win_ioctlsocket(SOCKET s, long cmd, u_long *argp) {

	rs_t rs;

	if (rs = rs_lookup(s))
		return rs_ioctl(rs->sd, cmd, argp);
	else
		return ioctlsocket(rs_lookup_socket(s), cmd, argp);
}

int __stdcall rs_win_setsockopt(SOCKET s, int level, int opt,
								char *optval, int optlen) {

	if (rs_lookup(s))
		return rs_setsockopt(rs_lookup_socket(s), level, opt, optval, optlen);
	else
		return setsockopt(rs_lookup_socket(s), level, opt, optval, optlen);
}

int __stdcall rs_win_WSAAsyncSelect(SOCKET s, HWND w, unsigned msg, long ev) {

	return add_async_select(s, w, msg, ev, rs_lookup(s));
}

int __stdcall rs_win_WSAGetLastError() {

	int i = WSAGetLastError();
	return i;
}

HMODULE __stdcall rs_win_LoadLibraryA(PCSTR module_path) {

	HMODULE mod = LoadLibraryA(module_path);
	if (mod != NULL)
		replace_rs_functions(mod);
	return mod;
}

HMODULE __stdcall rs_win_LoadLibraryW(PCWSTR mod_path) {

	HMODULE mod = LoadLibraryW(mod_path);
	if (mod != NULL)
		replace_rs_functions(mod);
	return mod;
}

HMODULE __stdcall rs_win_LoadLibraryExA(PCSTR mod_path, HANDLE file, DWORD flags) {

	HMODULE mod = LoadLibraryExA(mod_path, file, flags);
	if (mod != NULL && !(flags & LOAD_LIBRARY_AS_DATAFILE))
		replace_rs_functions(mod);
	return mod;
}

HMODULE __stdcall rs_win_LoadLibraryExW(PCWSTR mod_path, HANDLE file, DWORD flags) {

	HMODULE mod = LoadLibraryExW(mod_path, file, flags);
	if (mod != NULL && !(flags & LOAD_LIBRARY_AS_DATAFILE))
		replace_rs_functions(mod);
	return mod;
}

PROC __stdcall rs_win_GetProcAddress(HMODULE mod, PCSTR proc_name) {

	int i;
	DWORD *return_addr;
	PROC proc = GetProcAddress(mod, proc_name);
	
	/* check if call is from winsock */
	return_addr = (DWORD *)(&mod) - 1;			/* THIS MUST BE FIRST PARAMETER */
	if (*return_addr >= (DWORD)(winsock_dll.module) &&
			*return_addr < winsock_dll_end)
		return proc;
	
	for (i = 0; i < num_hooks; i++) {
		if (proc == hooks[i].old_proc)
			return hooks[i].new_proc;
	}

	return proc;
}

BOOL __stdcall rs_win_GetMessage(LPMSG msg, HWND wnd, UINT min, UINT max) {

	BOOL rv;

	rv = GetMessage(msg, wnd, min, max);
	event_show_message(msg);

	return rv;
}

void __stdcall die() {

	assert(0);
}

void replace_function(PCSTR old_module_name, PROC old_f, PROC new_f,
					  HMODULE caller_module) {

	ULONG size;
	PIMAGE_THUNK_DATA thunk;
	PIMAGE_IMPORT_DESCRIPTOR import_desc;

	/* get the import section */
	import_desc = (PIMAGE_IMPORT_DESCRIPTOR)
		ImageDirectoryEntryToData(caller_module, TRUE,
		                          IMAGE_DIRECTORY_ENTRY_IMPORT, &size);

	/* nothing to do if no imports */
	if (import_desc == NULL)
		return;

	/* find import descriptor for module in question */
	while (import_desc->Name) {
		PSTR mod_name = (PSTR)((PBYTE)caller_module + import_desc->Name);
		if (lstrcmpiA(mod_name, old_module_name) == 0)
			break;
		import_desc++;
	}
	if (import_desc->Name == 0)		/* do nothing if not found */
		return;

	thunk = (PIMAGE_THUNK_DATA)((PBYTE)caller_module + import_desc->FirstThunk);

	while (thunk->u1.Function) {

		PROC *pf = (PROC *)&thunk->u1.Function;

		if (*pf == old_f) {

			DWORD dummy;
			VirtualProtect(pf, sizeof(pf), PAGE_EXECUTE_READWRITE, &dummy); 
			assert(WriteProcessMemory(GetCurrentProcess(), pf, &new_f,
			                          sizeof(new_f), NULL));
			return;
		}

		thunk++;
	}
}

void replace_rs_functions(HMODULE caller_module) {

	int i;

	for (i = 0; i < num_hooks; i++)
		replace_function(hooks[i].old_module_name, hooks[i].old_proc,
		                 hooks[i].new_proc, caller_module);
}

void replace_all_rs_functions() {

	HANDLE snapshot;
	MODULEENTRY32 module_entry;
	MEMORY_BASIC_INFORMATION mbi;
	HMODULE rocks_module;
	BOOL good;

	VirtualQuery(replace_all_rs_functions, &mbi, sizeof(mbi));
	rocks_module = mbi.AllocationBase;
	snapshot = CreateToolhelp32Snapshot(TH32CS_SNAPMODULE, GetCurrentProcessId());
	module_entry.dwSize = sizeof(module_entry);
	good = Module32First(snapshot, &module_entry);
	while (good) {
		if (module_entry.hModule == rocks_module) {
			good = Module32Next(snapshot, &module_entry);
			continue;
		}
		replace_rs_functions(module_entry.hModule);
		good = Module32Next(snapshot, &module_entry);
	}

	CloseHandle(snapshot);
}

void add_hook(module_t mod, PCSTR name, PROC new_proc) {

	PROC old_proc;

	/* FIXME: do we need to be sure the library is loaded */
	assert(num_hooks < NUMHOOKS);
	old_proc = GetProcAddress(mod.module, name);
	if (old_proc == NULL)
		return;
	hooks[num_hooks].old_proc = old_proc;
	hooks[num_hooks].new_proc = new_proc;
	hooks[num_hooks].old_module_name = mod.module_name;
	++num_hooks;
}

int rs_init_sys() {

	MODULEINFO mod_info;
	DWORD tmp;

	/* get needed module handles */
	winsock_dll.module = GetModuleHandleA(winsock_dll.module_name);
	old_winsock_dll.module = GetModuleHandleA(old_winsock_dll.module_name);
	kernel_dll.module = GetModuleHandleA(kernel_dll.module_name);
	user_dll.module = GetModuleHandleA(user_dll.module_name);

	/* get size of winsock DLL */
	GetModuleInformation(GetCurrentProcess(), winsock_dll.module,
	                     &mod_info, sizeof(mod_info));
	tmp = (DWORD)(winsock_dll.module);
	winsock_dll_end = tmp + mod_info.SizeOfImage;

	/* build the list of functions to replace */
	add_hook(winsock_dll, "WSAStartup", rs_win_WSAStartup);
	add_hook(winsock_dll, "socket", (PROC)rs_win_socket);
	add_hook(winsock_dll, "connect", rs_win_connect);
	add_hook(winsock_dll, "closesocket", rs_win_closesocket);
	add_hook(winsock_dll, "recv", rs_win_recv);
	add_hook(winsock_dll, "send", rs_win_send);
	add_hook(winsock_dll, "accept", (PROC)rs_win_accept);
	add_hook(winsock_dll, "bind", rs_win_bind);
	add_hook(winsock_dll, "listen", rs_win_listen);
	add_hook(winsock_dll, "select", rs_win_select);
	add_hook(winsock_dll, "getsockname", rs_win_getsockname);
	add_hook(winsock_dll, "getpeername", rs_win_getpeername);
	add_hook(winsock_dll, "ioctlsocket", rs_win_ioctlsocket);
	add_hook(winsock_dll, "setsockopt", rs_win_setsockopt);
	add_hook(winsock_dll, "WSAAsyncSelect", rs_win_WSAAsyncSelect);
	add_hook(winsock_dll, "WSAGetLastError", rs_win_WSAGetLastError);
	add_hook(winsock_dll, "WSASend", (PROC)die);
	add_hook(winsock_dll, "WSARecv", (PROC)die);
	add_hook(old_winsock_dll, "WSAStartup", rs_win_WSAStartup);
	add_hook(old_winsock_dll, "socket", (PROC)rs_win_socket);
	add_hook(old_winsock_dll, "connect", rs_win_connect);
	add_hook(old_winsock_dll, "closesocket", rs_win_closesocket);
	add_hook(old_winsock_dll, "recv", rs_win_recv);
	add_hook(old_winsock_dll, "send", rs_win_send);
	add_hook(old_winsock_dll, "accept", (PROC)rs_win_accept);
	add_hook(old_winsock_dll, "bind", rs_win_bind);
	add_hook(old_winsock_dll, "listen", rs_win_listen);
	add_hook(old_winsock_dll, "select", rs_win_select);
	add_hook(old_winsock_dll, "getsockname", rs_win_getsockname);
	add_hook(old_winsock_dll, "getpeername", rs_win_getpeername);
	add_hook(old_winsock_dll, "ioctlsocket", rs_win_ioctlsocket);
	add_hook(old_winsock_dll, "setsockopt", rs_win_setsockopt);
	add_hook(old_winsock_dll, "WSAAsyncSelect", rs_win_WSAAsyncSelect);
	add_hook(old_winsock_dll, "WSAGetLastError", rs_win_WSAGetLastError);
	add_hook(kernel_dll, "LoadLibraryA", (PROC)rs_win_LoadLibraryA);
	add_hook(kernel_dll, "LoadLibraryW", (PROC)rs_win_LoadLibraryW);
	add_hook(kernel_dll, "LoadLibraryExA", (PROC)rs_win_LoadLibraryExA);
	add_hook(kernel_dll, "LoadLibraryExW", (PROC)rs_win_LoadLibraryExW);
	add_hook(kernel_dll, "GetProcAddress", (PROC)rs_win_GetProcAddress);


	add_hook(user_dll, "GetMessageA", (PROC)rs_win_GetMessage);

	replace_all_rs_functions();

	return 0;
}