/* 
 *  rocks/test/sock.c
 *
 *  Socket test program.
 *
 *  Copyright (C) 2001 Victor Zandy
 *  See COPYING for distribution terms.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <netdb.h>
#include <sys/time.h>
#include <assert.h>
#include <errno.h>

/* FIXME: Note that we use stdout, not stderr, for diagnostics.  This
   way they show up on the terminal, not in the rock log. */

static int DEBUG = 0;

/* Individual runs for measuring connection time. */
#define MAXRUNS            10000

#define DEFAULT_BUFFER_SIZE   (8*1024*1024)
static unsigned char *buffer1;
static unsigned char *buffer2;

#define MIN(x,y) ((x) < (y) ? (x) : (y))


typedef enum { ONE_WAY_WRITE,
	       TWO_WAY_WRITE,
	       CONNECT_AND_DROP,
	       ACCEPT_AND_DROP,
	       READ_AND_DISCARD,
	       PING,
	       IDLE,
	       STD } smode_t;


static double
mbps(unsigned bytes, struct timeval *tv)
{
	double m;
	m = bytes * 8;                           /* bits */
	m /= (tv->tv_sec*1000000 + tv->tv_usec); /* bits/usec */
	m *= 1000000;                            /* bits/sec  */
	m /= (1024 * 1024);                      /* Mbits/sec */
	return m;
}

static struct in_addr
hostname_to_addr(char *hostname)
{
	struct hostent* h;
	struct in_addr addr;
	h = gethostbyname(hostname);
	if (!h) {
		fprintf(stdout, "Host lookup failed: %s\n", hostname);
		exit(1);
	}
	addr = *((struct in_addr *) h->h_addr); /* network order */
	return addr;
}

static int
xread(int sd, void *buf, size_t len)
{
	char *p = (char *)buf;
	size_t nrecv = 0;
	ssize_t rv;
	
	while (nrecv < len) {
		rv = read(sd, p, len - nrecv);
		if (0 > rv && errno == EINTR)
			continue;
		if (0 > rv)
			return -1;
		if (0 == rv)
			return 0;
		nrecv += rv;
		p += rv;
	}
	return nrecv;
}

static int
xwrite(int sd, void *buf, size_t len)
{
	char *p = (char *)buf;
	size_t nsent = 0;
	ssize_t rv;
	
	while (nsent < len) {
		rv = write(sd, p, len - nsent);
		if (0 > rv && errno == EINTR)
			continue;
		if (0 > rv)
			return -1;
		nsent += rv;
		p += rv;
	}
	return nsent;
}

static
int tv_cmp(const struct timeval *a,
	   const struct timeval *b)
{
	if (a->tv_sec < b->tv_sec)
		return -1;
	if (a->tv_sec > b->tv_sec)
		return 1;
	if (a->tv_usec < b->tv_usec)
		return -1;
	if (a->tv_usec > b->tv_usec)
		return 1;
	return 0;
}

/* c = a - b */
static
void tv_diff(const struct timeval *a,
	     const struct timeval *b,
	     struct timeval *c)
{
	c->tv_sec = a->tv_sec - b->tv_sec;
	c->tv_usec = a->tv_usec - b->tv_usec;
	if (c->tv_usec < 0) {
		c->tv_sec -= 1;
		c->tv_usec += 1000000;
	}
}

static
void tv_add(const struct timeval *a,
	    const struct timeval *b,
	    struct timeval *c)
{
	c->tv_sec = a->tv_sec + b->tv_sec;
	c->tv_usec = a->tv_usec + b->tv_usec;
	if (c->tv_usec > 1000000) {
		c->tv_sec += 1;
		c->tv_usec -= 1000000;
	}
}

static int
do_discard(int sock, int buflen, struct timeval *tv)
{
	int rv;
	unsigned nread;
	unsigned loops;
	struct timeval s;

	if (tv) {
		fd_set fds;
		/* Wait for data to be ready */
		FD_ZERO(&fds);
		FD_SET(sock, &fds);
		if (0 > select(sock+1, &fds, NULL, NULL, NULL)) {
			perror("select");
			return -1;
		}
		gettimeofday(&s, NULL);
	}
	nread = 0;
	loops = 0;
	while (1) {
		rv = read(sock, buffer1, buflen);
		if (0 > rv) {
			perror("read");
			return 0;
		}
		if (0 == rv) {
			if (DEBUG) {
				fprintf(stdout, "EOF\n");
				fflush(stdout);
			}
			goto out;
		}
		nread += rv;
		loops++;
		if (DEBUG) {
			fprintf(stdout, ".");
			fflush(stdout);
		}
	}
 out:
	if (tv && nread > 0) {
		gettimeofday(tv, NULL);
		tv_diff(tv, &s, tv);
		printf(" %6.1f Mbits/sec %6.0f bytes/read %d bytes total\n",
		       mbps(nread, tv),
		       (double)nread/loops,
		       nread);
		fflush(stdout);
	}
	return 0;
}


/* Connect to HOST:PORT and return the new socket.  If TV is non-null,
   return in TV the wall time for establishing the connection. */
static int
do_connect(char *host, short port, struct timeval *tv)
{
	int sd;
	struct sockaddr_in addr;
	int len;
	struct timeval s, e;
	int rv, ern;

	sd = socket(AF_INET, SOCK_STREAM, 0);
	if (0 > sd) {
		perror("socket");
		return -1;
	}
	bzero(&addr, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr = hostname_to_addr(host);
	addr.sin_port = htons(port);
	len = sizeof(addr);
	if (tv)
		gettimeofday(&s, NULL);
	rv = connect(sd, (struct sockaddr *)&addr, len);
	ern = errno;
	if (tv) {
		gettimeofday(&e, NULL);
		tv_diff(&e, &s, tv);
	}
	if (0 > rv) {
		errno = ern;
		perror("connect");
		return -1;
	}
	return sd;
}

/* Connect to HOST:PORT N times, and close the connection after each
   connect.  If TV is non-null, return in TV the total wall time for
   all the connects.  Return 0 on success, -1 otherwise. */
static int
do_many_connects(char *host, short port, int N, struct timeval *tv)
{
	struct sockaddr_in addr;
	int len;
	int sock;
	struct timeval s, e;
	int rv, ern;

	bzero(&addr, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr = hostname_to_addr(host);
	addr.sin_port = htons(port);
	len = sizeof(addr);
	
	if (tv)
		bzero(tv, sizeof(struct timeval));

	if (tv)
		gettimeofday(&s, NULL);
	while (N != 0) {
		sock = socket(AF_INET, SOCK_STREAM, 0);
		if (0 > sock) {
			perror("socket");
			return -1;
		}
		rv = connect(sock, (struct sockaddr *)&addr, len);
		ern = errno;
		if (0 > rv) {
			errno = ern;
			perror("connect");
			return -1;
		}
		close(sock);
		if (DEBUG) {
			fprintf(stdout, ".");
			fflush(stdout);
		}
		if (N > 0)
			N--;
	}	
	if (tv) {
		gettimeofday(&e, NULL);
		tv_diff(&e, &s, tv);
	}
	return 0;
}

/* PORT is in network order */
static int
do_lb(short port, struct sockaddr_in *addr)
{
	int sd;
	int len;

	sd = socket(AF_INET, SOCK_STREAM, 0);
	if (0 > sd) {
		perror("socket");
		return -1;
	}

	bzero(addr, sizeof(struct sockaddr_in));
	addr->sin_family = AF_INET;
	addr->sin_addr.s_addr = INADDR_ANY;
	addr->sin_port = port;
	len = sizeof(struct sockaddr_in);
	if (0 > bind(sd, (struct sockaddr *) addr, len)) {
		perror("bind");
		return -1;
	}
	len = sizeof(addr);
	if (0 > getsockname(sd, (struct sockaddr *) addr, &len)) {
		perror("getsockname");
		return -1;
	}
	if (0 > listen(sd, 1)) {
		perror("listen");
		return -1;
	}

	fprintf(stderr,
		"Server listening on port %d\n", ntohs(addr->sin_port));
	return sd;
}


static int
do_accept(int serv)
{
	int sock;
	struct sockaddr_in addr;
	int len;

	len = sizeof(addr);
	sock = accept(serv, (struct sockaddr *) &addr, &len);
	if (0 > sock) {
		perror("accept");
		return -1;
	}
	return sock;
}


static int
do_lba(int closeserv)
{
	int sd, cd;
	struct sockaddr_in addr;
	int len;

	sd = socket(AF_INET, SOCK_STREAM, 0);
	if (0 > sd) {
		perror("socket");
		return -1;
	}

	bzero(&addr, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = htons(0);
	len = sizeof(addr);
	if (0 > bind(sd, (struct sockaddr *) &addr, len)) {
		perror("bind");
		return -1;
	}
	len = sizeof(addr);
	if (0 > getsockname(sd, (struct sockaddr *) &addr, &len)) {
		perror("getsockname");
		return -1;
	}
	if (0 > listen(sd, 10)) {
		perror("listen");
		return -1;
	}

	fprintf(stderr, "Server listening on port %d\n", ntohs(addr.sin_port));

	len = sizeof(addr);
	cd = accept(sd, (struct sockaddr *) &addr, &len);
	if (0 > cd) {
		perror("accept");
		return -1;
	}
	if (closeserv && 0 > close(sd)) {
		perror("close");
		return -1;
	}
	return cd;
}

static int
do_many_lba(int n)
{
	int sd, cd;
	struct sockaddr_in addr;
	int len;

	sd = socket(AF_INET, SOCK_STREAM, 0);
	if (0 > sd) {
		perror("socket");
		return -1;
	}

	bzero(&addr, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = htons(0);
	len = sizeof(addr);
	if (0 > bind(sd, (struct sockaddr *) &addr, len)) {
		perror("bind");
		return -1;
	}
	len = sizeof(addr);
	if (0 > getsockname(sd, (struct sockaddr *) &addr, &len)) {
		perror("getsockname");
		return -1;
	}
	if (0 > listen(sd, 10)) {
		perror("listen");
		return -1;
	}

	fprintf(stderr, "Server listening on port %d\n", ntohs(addr.sin_port));

	while (n != 0) {
		len = sizeof(addr);
		cd = accept(sd, (struct sockaddr *) &addr, &len);
		if (0 > cd) {
			perror("accept");
			return -1;
		}
		close(cd);
		if (DEBUG) {
			fprintf(stdout, ".");
			fflush(stdout);
		}
		if (n > 0)
			n--;
	}
	return 0;
}

/* Non-negative N -> write N bytes total in BUFLEN chunks.
   Negative N -> write BUFLEN bytes until the cows come home. */
static int
do_one_way_write(int sock, const int buflen, int n, struct timeval *tv)
{
	unsigned long sent, recd;
	struct timeval s, e;

	sent = recd = 0;
	if (tv)
		gettimeofday(&s, NULL);
	while (n != 0) {
		if (0 > xwrite(sock, buffer1,
			       n > 0 ? MIN(buflen, n) : buflen)) {
			perror("write");
			return -1;
		}
		sent += buflen;
		if (DEBUG) {
			fprintf(stdout, ".");
			fflush(stdout);
		}
		if (n > 0)
			n -= buflen;
	}
	if (tv) {
		gettimeofday(&e, NULL);
		tv_diff(&e, &s, tv);
	}
	return 0;
}

static int
do_ping_server(int sock, int n, struct timeval *tv)
{
	unsigned long len;
	struct timeval s, e;
	int rv;

	/* Get packet size */
	rv = xread(sock, &len, sizeof(len));
	if (0 > rv)
		return -1;
	if (0 == rv)
		return 0;
	len = ntohl(len);

	if (tv)
		gettimeofday(&s, NULL);
	while (n != 0) {
		int rv;
		rv = xread(sock, buffer1, len);
		if (0 > rv) {
			perror("read");
			return -1;
		}
		if (0 == rv) {
			if (DEBUG) {
				fprintf(stdout, "EOF\n");
				fflush(stdout);
			}
			return 0;
		}
		if (0 > xwrite(sock, buffer1, len)) {
			perror("write");
			return -1;
		}
		if (DEBUG) {
			fprintf(stdout, ".");
			fflush(stdout);
		}
		if (n > 0)
			n--;
	}
	if (tv) {
		gettimeofday(&e, NULL);
		tv_diff(&e, &s, tv);
	}
	return 0;
}


static int
do_ping_client(int sock, const int buflen, int n, struct timeval *tv)
{
	struct timeval s, e;
	unsigned long len;

	/* Tell server how big the packets will be */
	len = htonl(buflen);
	if (0 > xwrite(sock, &len, sizeof(len)))
		return -1;
	sleep(1); /* Wait for the dust to settle */
	if (tv)
		gettimeofday(&s, NULL);
	while (n != 0) {
		if (0 > xwrite(sock, buffer1, buflen)) {
			perror("read");
			return -1;
		}
		if (0 > xread(sock, buffer1, buflen)) {
			perror("write");
			return -1;
		}
		if (DEBUG) {
			fprintf(stdout, ".");
			fflush(stdout);
		}
		if (n > 0)
			n--;
	}
	if (tv) {
		gettimeofday(&e, NULL);
		tv_diff(&e, &s, tv);
	}
	return 0;
}

/* Client send BUFLEN bytes, then wait for a 1 byte ack from the
   server.  Server receive BUFLEN bytes, then reply 1 byte ack. */
static int
do_ping(int sock, const int buflen, int n, int server, struct timeval *tv)
{
	if (server)
		return do_ping_server(sock, n, tv);
	else
		return do_ping_client(sock, buflen, n, tv);
}

/* Exchange messages of size BUFLEN over SOCK.
   Indicate an error if the message received differs from
   the message sent.
   N < 0 => run forever
   N > 0 => exchange N messages
   Return 0 on success, -1 on error. */
static int
do_two_way_write(int sock, const int buflen, int n)
{
	unsigned long i;
	unsigned long sent, recd;

	for (i = 0; i < buflen; i++)
		buffer1[i] = i % 256;

	sent = recd = 0;
	while (n != 0) {
		if (0 > xwrite(sock, buffer1, buflen)) {
			perror("write");
			return -1;
		}
		sent += buflen;
		if (0 > xread(sock, buffer2, buflen)) {
			perror("read");
			return -1;
		}
		recd += buflen;
		if (memcmp(buffer1, buffer2, buflen)) {
			int i;
			for (i = 0; i < buflen; i++)
				if (buffer1[i] != buffer2[i])
					break;
			fprintf(stdout,
				"Byte mismatch at sequence %ld\n", recd + i);
			return -1;
		}
		if (DEBUG) {
			fprintf(stdout, ".");
			fflush(stdout);
		}
		if (n > 0)
			n--;
	}
	return 0;
}

static int
do_idle()
{
	while (1)
		pause();
}

static unsigned long
parsebytes(char *p)
{
	unsigned long l, mul;
	char *q;

	/* 100, 100K, 100k, 100M, 100m */
	l = strtoul(p, &q, 10);
	if (!strlen(q))
		return l;
	if (strlen(q) > 1)
		goto err;
	if (*q == 'k' || *q == 'K')
		mul = 1024;
	else if (*q == 'm' || *q == 'M')
		mul = 1024 * 1024;
	else
		goto err;
	return l * mul;
 err:
	fprintf(stderr, "Bad byte count specification \"%s\", using %ld\n", p, l);
	return l;
}

static int
do_udp_server(smode_t mode, char *host, int port,
	      int sz, int n, int timing, int noiop)
{
	int sock;
	struct sockaddr_in addr;
	socklen_t len;
	int rv;
	struct timeval s, e, tv;
	int needtime;
	int todo;
	int npkts;

	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (0 > sock) {
		perror("socket");
		return -1;
	}
	bzero(&addr, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = htons(port);
	if (0 > bind(sock, (struct sockaddr *)&addr, sizeof(addr))) {
		perror("bind");
		return -1;
	}
	needtime = 1; /* Go one iteration before starting timer 
			 to allow rocks to connect. */
	assert(PING == mode || READ_AND_DISCARD == mode);
	todo = n;
	npkts = 0;
	while (todo > 0) {
		len = sizeof(addr);
		rv = recvfrom(sock, buffer1, sz, 0, (struct sockaddr *)&addr, &len);
		if (1 == rv)
			break;
		assert(rv == sz);
		npkts++;
		if (0 > rv) {
			perror("recvfrom");
			return -1;
		}
		if (PING == mode) {
			len = sizeof(addr);
			rv = sendto(sock, buffer1, sz, 0, (struct sockaddr *)&addr, len);
			if (0 > rv) {
				perror("sendto");
				return -1;
			}
			todo--;
		}
		if (DEBUG) {
			fprintf(stdout, ".");
			fflush(stdout);
		}
		if (timing && needtime) {
			gettimeofday(&s, NULL);
			needtime = 0;
		}
	}
	if (DEBUG)
		fprintf(stdout, "\n");
	if (timing) {
		gettimeofday(&e, NULL);
		tv_diff(&e, &s, &tv);
		if (READ_AND_DISCARD == mode)
			printf("received %d pkts (%d bytes) at %6.1f Mb/sec\n",
			       npkts, npkts*sz, mbps(npkts*sz, &tv));
	}
	close(sock);
	return 0;
}

static int
do_udp_client(smode_t mode, char *host, int port,
	      int sz, int n, int timing, int noiop)
{
	int sock;
	struct sockaddr_in addr, addr2;
	socklen_t len;
	int rv;
	struct timeval s, e, tv;
	int needtime;
	int todo;

	sock = socket(AF_INET, SOCK_DGRAM, 0);
	if (0 > sock) {
		perror("socket");
		return -1;
	}
	bzero(&addr, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr = hostname_to_addr(host);
	addr.sin_port = htons(port);
	needtime = 1; /* Go one iteration before starting timer 
			 to allow rocks to connect. */
	assert(PING == mode || ONE_WAY_WRITE == mode);
	todo = n;
	sleep(1);
	while (todo > 0) {
		len = sizeof(addr);
		rv = sendto(sock, buffer1, sz, 0, (struct sockaddr *)&addr, len);
		if (0 > rv) {
			perror("sendto");
			return -1;
		}
		if (PING == mode) {
			len = sizeof(addr2);
			rv = recvfrom(sock, buffer1, sz, 0, (struct sockaddr *)&addr2, &len);
			if (0 > rv) {
				perror("recvfrom");
				return -1;
			}
			assert(rv == sz);
			todo--;
		} else {
			todo -= rv;
		}
		if (DEBUG) {
			fprintf(stdout, ".");
			fflush(stdout);
		}
		if (timing && needtime) {
			gettimeofday(&s, NULL);
			needtime = 0;
		}
	}
	if (ONE_WAY_WRITE == mode) {
		/* Send "EOF" */
		rv = sendto(sock, buffer1, 1, 0, (struct sockaddr *)&addr, len);
		if (0 > rv) {
			perror("sendto");
			return -1;
		}
	}

	if (DEBUG)
		fprintf(stdout, "\n");
	if (timing) {
		unsigned long usec;
		gettimeofday(&e, NULL);
		tv_diff(&e, &s, &tv);
		usec = tv.tv_sec*1000000+tv.tv_usec;
		if (ONE_WAY_WRITE == mode)
			printf("%10d => %10ld usec, %6.1f Mb/sec\n",
			       sz, usec, mbps(n, &tv));
		else
			printf("%10d => %10ld usec, round-trip avg %6.0f usec\n",
			       sz, usec, (double)usec/(double)n);
	}
	close(sock);
	return 0;
}


static int
do_udp(smode_t mode, int server, char *host, int port,
       int sz, int n, int timing, int noiop)
{
	if (-1 == n) {
		fprintf(stdout, "Mode limits are required with UDP.\n");
		return -1;
	}
	if (timing && n < 2) {
		fprintf(stdout, "Warning: timing is bogus for 1 UDP iteration\n");
		timing = 0;
	}
	if (!port) {
		fprintf(stdout, "A port is required for all UDP modes.\n");
		return -1;
	}
	if (!(PING == mode
	      || (server && READ_AND_DISCARD == mode)
	      || (!server && ONE_WAY_WRITE == mode))) {
		fprintf(stdout,
			"UDP is only supported for read-and-discard, ping, and one-way-write\n");
		return -1;
	}
	if (server)
		return do_udp_server(mode, host, port,
				     sz, n, timing, noiop);
	else
		return do_udp_client(mode, host, port,
				     sz, n, timing, noiop);
}

static void
usage_and_exit(int e)
{
        fprintf(stdout, "Usage: sock [switches] [mode]\n");
        fprintf(stdout, " Switches:\n");
        fprintf(stdout, "  -s             Run as a server\n");
        fprintf(stdout, "  -h <host>      (Clients) Server host (default localhost)\n");
        fprintf(stdout, "  -p <port>      (Clients) Server port\n");
        fprintf(stdout, "  -d             Print debugging information\n");
        fprintf(stdout, "  -t             Print timing statistics\n");
        fprintf(stdout, "  -z <size>      Set network I/O buffer size\n");
	fprintf(stdout, "  -u             Use UDP\n");
        fprintf(stdout, "  -I             Disable rock detection support\n");
        fprintf(stdout, " Modes:\n");
        fprintf(stdout, "  -i             Just idle after establishing connection\n");
        fprintf(stdout, "  -1[n]          Do 1-way write of n bytes\n");
        fprintf(stdout, "  -2[n]          Do 2-way write n times\n");
        fprintf(stdout, "  -x[n]          Read and discard (n times for UDP)\n");
        fprintf(stdout, "  -a[n]          Just accept n times (servers only)\n");
        fprintf(stdout, "  -c[n]          Just connect n times (clients only)\n");
	fprintf(stdout, "  -l[n]          Send (clients) or receive (servers) n pings\n");
	fprintf(stdout, " If optional count n is omitted, do forever.\n");
	exit(e);
}

int
main(int argc, char *argv[])
{
	int c;
	int sock, serv;
	struct timeval tv;
	struct sockaddr_in addr;

	char *host = "localhost";      /* Host for client connections */
	int port = 0;                  /* Port for client connections */
	smode_t mode = STD;             /* Mode of operation */ 
	int n = -1;                    /* Iterations; -1 means loop forever */
	int sz = DEFAULT_BUFFER_SIZE;  /* Buffer size */
	int server = 0;                /* Server or client? */
	int timing = 0;                /* Collect timing statistics? */
	int noiop = 0;                 /* Disable interoperability compatibility? */
	int udp = 0;		       /* Use UDP */

	opterr = 1;
	optind = 0;
	while (EOF != (c = getopt(argc, argv, "sp:h:dtz:Iui1::2::c::a::l::x::")))
		switch (c) {
		case 's':
			server = 1;
			break;
		case 'p':
			port = atoi(optarg);
			break;
		case 'h':
			host = optarg;
			break;
		case 'd':
			DEBUG++;
			break;
		case 't':
			timing = 1;
			break;
		case 'z':
			sz = parsebytes(optarg);
			break;
		case 'i':
			mode = IDLE;
			break;
		case 'I':
			noiop = 1;
			break;
		case 'u':
			udp = 1;
			break;
		case '1':
			mode = ONE_WAY_WRITE;
			if (optarg)
				n = parsebytes(optarg);
			else
				n = -1;
			break;
		case '2':
			mode = TWO_WAY_WRITE;
			if (optarg)
				n = parsebytes(optarg);
			else
				n = -1;
			break;
		case 'a':
			mode = ACCEPT_AND_DROP;
			server = 1;
			if (optarg)
				n = parsebytes(optarg);
			else
				n = -1;
			break;
		case 'c':
			mode = CONNECT_AND_DROP;
			server = 0;
			if (optarg)
				n = parsebytes(optarg);
			else
				n = -1;
			break;
		case 'l':
			mode = PING;
			if (optarg)
				n = parsebytes(optarg);
			else
				n = -1;
			break;
		case 'x':
			mode = READ_AND_DISCARD;
			if (optarg)
				n = parsebytes(optarg);
			else
				n = -1;
			break;
		case '?':
			fprintf(stdout, "Unknown option `-%c'.\n", c);
			usage_and_exit(1);
			break;
		}
	if (!server && (!host || !port))
		usage_and_exit(1);

	buffer1 = (unsigned char*) malloc(sz);
	buffer2 = (unsigned char*) malloc(sz);
	if (!buffer1 || !buffer2) {
		fprintf(stdout, "Out of memory.\n");
		exit(1);
	}
	if (udp)
		return do_udp(mode, server, host, port, sz, n, timing, noiop);

	/* Initialize buffers to avoid paging effects later. */
	bzero(buffer1, sz);
	bzero(buffer2, sz);

	if (timing
	    && !server
	    && CONNECT_AND_DROP == mode
	    && (n < 0 || n > MAXRUNS)) {
		fprintf(stdout, "Too many connect and drop runs to time.\n");
		exit(1);
	}

	if (server)
		switch (mode) {
		case IDLE:
			sock = do_lba(0);
			if (0 > sock) {
				fprintf(stderr, "Failed to create server\n");
				exit(1);
			}
			do_idle();
			break;
		case READ_AND_DISCARD:
			serv = do_lb(htons(0), &addr);
			if (0 > serv) {
				fprintf(stdout, "Failed to create listener\n");
				exit(1);
			}
			while (1) {
				sock = do_accept(serv);
				if (noiop)
					close(serv);
				if (0 > sock) {
					fprintf(stdout, "Failed to accept client\n");
					exit(1);
				}
				if (0 > do_discard(sock, sz, timing ? &tv : NULL)) {
					fprintf(stdout, "Server loop failed\n");
					exit(1);
				}
				close(sock);
				if (noiop) {
					/* Reuse previous listener port. */
					serv = do_lb(addr.sin_port, &addr);
					if (0 > serv) {
						fprintf(stdout, "Failed to create listener\n");
						exit(1);
					}
				}
			}
			break;
		case ONE_WAY_WRITE:
			sock = do_lba(0);
			if (0 > sock) {
				fprintf(stdout, "Failed to create server\n");
				exit(1);
			}
			if (0 > do_one_way_write(sock, sz, n, timing ? &tv : NULL)) {
				fprintf(stdout, "Server loop failed\n");
				exit(1);
			}
			close(sock);
			break;
		case PING:
			serv = do_lb(htons(0), &addr);
			if (0 > serv) {
				fprintf(stdout, "Failed to create listener\n");
				exit(1);
			}
			while (1) {
				sock = do_accept(serv);
				if (noiop)
					close(serv);
				if (0 > sock) {
					fprintf(stdout, "Failed to accept client\n");
					exit(1);
				}
				if (0 > do_ping_server(sock, n, timing ? &tv : NULL)) {
					fprintf(stdout, "Ping server loop failed\n");
					exit(1);
				}
				close(sock);
				if (noiop) {
					/* Reuse previous listener port. */
					serv = do_lb(addr.sin_port, &addr);
					if (0 > serv) {
						fprintf(stdout,
							"Failed to create listener\n");
						exit(1);
					}
				}
			}
			break;
		case TWO_WAY_WRITE:
			serv = do_lb(htons(0), &addr);
			if (0 > serv) {
				fprintf(stdout, "Failed to create server\n");
				exit(1);
			}
			while (1) {
				sock = do_accept(serv);
				if (noiop)
					close(serv);
				if (0 > sock) {
					fprintf(stdout, "Failed to accept client\n");
					exit(1);
				}
				if (0 > do_two_way_write(sock, sz, n)) {
					fprintf(stdout, "Server loop failed\n");
					exit(1);
				}
				close(sock);
				if (noiop) {
					/* Reuse previous listener port. */
					serv = do_lb(addr.sin_port, &addr);
					if (0 > serv) {
						fprintf(stdout, "Failed to create listener\n");
						exit(1);
					}
				}
			}
			break;
		case CONNECT_AND_DROP:
			usage_and_exit(1);
			break;
		case ACCEPT_AND_DROP:
			if (0 > do_many_lba(n)) {
				fprintf(stdout, "Server loop failed\n");
				exit(1);
			}
			break;
		case STD:
			fprintf(stdout, "Standard usage not yet supported\n");
			exit(1);
			break;
		default:
			usage_and_exit(1);			
		}
	else
		switch (mode) {
		case IDLE:
			sock = do_connect(host, port, timing ? &tv : NULL);
			if (0 > sock) {
				fprintf(stdout, "Client connect failed\n");
				exit(1);
			}
			do_idle();
			break;
		case ONE_WAY_WRITE:
			sock = do_connect(host, port, timing ? &tv : NULL);
			if (0 > sock) {
				fprintf(stdout, "Client connect failed\n");
				exit(1);
			}
			if (timing)
				sleep(1);
			if (0 > do_one_way_write(sock, sz, n,
						 timing ? &tv : NULL)) {
				fprintf(stdout, "Client loop failed\n");
				exit(1);
			}
			close(sock);
			if (timing) {
				printf("%10d => %10ld usec, %6.1f Mb/sec\n",
				       sz,
				       tv.tv_sec*1000000+tv.tv_usec,
				       mbps(n, &tv));
			}
			break;
		case PING:
			sock = do_connect(host, port, NULL);
			if (0 > sock) {
				fprintf(stdout, "Client connect failed\n");
				exit(1);
			}
			if (timing)
				/* Be sure server is ready. */
				sleep(1);
			if (0 > do_ping_client(sock, sz, n,
					       timing ? &tv : NULL)) {
				fprintf(stdout, "Ping client loop failed\n");
				exit(1);
			}
			close(sock);
			if (timing) {
				unsigned long usec;
				usec = tv.tv_sec*1000000+tv.tv_usec;
				printf("%10d => %10ld usec, round-trip avg %6.0f usec\n",
				       sz, usec, (double)usec/(double)n);
			}
			break;
		case TWO_WAY_WRITE:
			sock = do_connect(host, port, timing ? &tv : NULL);
			if (0 > sock) {
				fprintf(stdout, "Client connect failed\n");
				exit(1);
			}
			if (0 > do_two_way_write(sock, sz, n)) {
				fprintf(stdout, "Client loop failed\n");
				exit(1);
			}
			close(sock);
			break;
		case CONNECT_AND_DROP:
			if (0 > do_many_connects(host, port,
						 n, timing ? &tv : NULL)) {
				fprintf(stdout, "Client connect failed\n");
				exit(1);
			}
			if (timing) {
				long usec;
				usec = tv.tv_sec*1000000 + tv.tv_usec;
				fprintf(stdout, "=> %6.2f usec/connect, %ld usec total, %d connects\n",
					usec/(double)n, usec, n);
			}
			break;
		case ACCEPT_AND_DROP:
			usage_and_exit(1);
			break;
		case READ_AND_DISCARD:
			/* Accept a connection on the socket.
			   Read from it until it closes.
			   Repeat. */
			sock = do_connect(host, port, timing ? &tv : NULL);
			if (0 > sock) {
				fprintf(stdout, "Client connect failed\n");
				exit(1);
			}
			if (0 > do_discard(sock, sz, timing ? &tv : NULL)) {
				fprintf(stdout, "Client loop failed\n");
				exit(1);
			}
			close(sock);
			break;
		case STD:
			fprintf(stdout, "Standard usage not yet supported\n");
			exit(1);
			break;
		default:
			usage_and_exit(1);
		}
	free(buffer1);
	free(buffer2);
	return 0;
}
