/* 
 *  rocks/test/test_inflight.c
 *
 *  In flight data test program.
 *
 *  Copyright (C) 2001 Victor Zandy
 *  See COPYING for distribution terms.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <netdb.h>
#include <sys/time.h>
#include <assert.h>
#include <errno.h>

static int opt_serv = 0;

static struct in_addr
hostname_to_addr(char *hostname)
{
	struct hostent* h;
	struct in_addr addr;
	h = gethostbyname(hostname);
	if (!h) {
		fprintf(stderr, "Host lookup failed: %s\n", hostname);
		exit(1);
	}
	addr = *((struct in_addr *) h->h_addr); /* network order */
	return addr;
}

static int
xread(int sd, void *buf, size_t len)
{
	char *p = (char *)buf;
	size_t nrecv = 0;
	ssize_t rv;
	
	while (nrecv < len) {
		rv = read(sd, p, len - nrecv);
		if (0 > rv && errno == EINTR)
			continue;
		if (0 > rv)
			return -1;
		nrecv += rv;
		p += rv;
	}
	return nrecv;
}

static int
xwrite(int sd, void *buf, size_t len)
{
	char *p = (char *)buf;
	size_t nsent = 0;
	ssize_t rv;
	
	while (nsent < len) {
		rv = write(sd, p, len - nsent);
		if (0 > rv && errno == EINTR)
			continue;
		if (0 > rv)
			return -1;
		nsent += rv;
		p += rv;
	}
	return nsent;
}

static int
clnt(char *host, short port)
{
	int sd;
	struct sockaddr_in addr;
	int len;

	sd = socket(AF_INET, SOCK_STREAM, 0);
	if (0 > sd) {
		perror("socket");
		return -1;
	}
	bzero(&addr, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr = hostname_to_addr(host);
	addr.sin_port = htons(port);
	len = sizeof(addr);
	if (0 > connect(sd, &addr, len)) {
		perror("connect");
		return -1;
	}
	fprintf(stderr, "Connected.\n");
	return sd;
}

static int
serv()
{
	int sd, cd;
	struct sockaddr_in addr;
	int len;

	sd = socket(AF_INET, SOCK_STREAM, 0);
	if (0 > sd) {
		perror("socket");
		return -1;
	}

	bzero(&addr, sizeof(addr));
	addr.sin_family = AF_INET;
	addr.sin_addr.s_addr = INADDR_ANY;
	addr.sin_port = htons(0);
	len = sizeof(addr);
	if (0 > bind(sd, (struct sockaddr *) &addr, len)) {
		perror("bind");
		return -1;
	}
	len = sizeof(addr);
	if (0 > getsockname(sd, (struct sockaddr *) &addr, &len)) {
		perror("getsockname");
		return -1;
	}
	if (0 > listen(sd, 1)) {
		perror("listen");
		return -1;
	}

	fprintf(stderr, "Server listening on port %d\n", ntohs(addr.sin_port));

	len = sizeof(addr);
	cd = accept(sd, (struct sockaddr *) &addr, &len);
	if (0 > cd) {
		perror("accept");
		return -1;
	}
	if (0 > close(sd)) {
		perror("close");
		return -1;
	}
	return cd;
}

#define BUFSIZE 10000
static unsigned long buf[BUFSIZE];
static unsigned long orig[BUFSIZE];

static int
flight_test(int sock)
{
	int i = 0;
	unsigned long sent, recd;

	for (i = 0; i < BUFSIZE; i++)
		orig[i] = htonl(i);

	sent = recd = 0;
	while (1) {
		if (0 > xwrite(sock, orig, sizeof(orig))) {
			perror("write");
			return -1;
		}
		sent += sizeof(orig);
		if (0 > xread(sock, buf, sizeof(buf))) {
			perror("read");
			return -1;
		}
		recd += sizeof(buf);
		if (memcmp(buf, orig, sizeof(buf))) {
			int i;
			for (i = 0; i < sizeof(buf); i++)
				if (buf[i] != orig[i])
					break;
			fprintf(stderr, "Byte mismatch at sequence %ld\n", recd + i);
			return -1;
		}
		fprintf(stdout, "."); fflush(stdout);
	}
}

int
main(int argc, char *argv[])
{
	int c;
	int sock;
	int port = 0;
	char *host = NULL;

	opterr = 1;
	optind = 0;
	while (EOF != (c = getopt(argc, argv, "sh:p:")))
		switch (c) {
		case 's':
			opt_serv++;
			break;
		case 'p':
			port = atoi(optarg);
			break;
		case 'h':
			host = optarg;
			break;
		case '?':
			fprintf(stderr, "Unknown option `-%c'.\n", c);
			exit(1);
			break;
		}
	if (!opt_serv && (!host || !port)) {
		fprintf(stderr, "Usage:  %s -h HOST -p PORT  (client)\n", argv[0]);
		fprintf(stderr, "Usage:  %s -s               (server)\n", argv[0]);
		exit(1);
	}
	sock = opt_serv ? serv() : clnt(host, port);
	if (0 > sock) {
		fprintf(stderr, "Failed to start test_inflight %s\n",
			opt_serv ? "server" : "client");
		return -1;
	}
	return flight_test(sock);
}
