/* 
 *  rocks/util.c
 *
 *  Common utility functions.
 *
 *  Copyright (C) 2001 Victor Zandy
 *  See COPYING for distribution terms.
 */

#include "log.h"
#include "rs.h"

#include <string.h>
#include <errno.h>
#include <assert.h>
#include <stdio.h>

int
rs_xwrite(socket_t sd, iobuff_t buf, iosize_t len)
{
	char *p = (char *)buf;
	iosize_t nsent = 0;
	ioret_t rv;

	while (nsent < len) {
		rv = send(sd, p, len - nsent, 0);
		if (0 > rv && (errno == EINTR || errno == EWOULDBLOCK))
			continue;
		if (0 > rv)
			return -1;
		nsent += rv;
		p += rv;
	}
	return nsent;
}

/* c = a-b */
void
rs_tv_diff(const struct timeval *a,
	   const struct timeval *b,
	   struct timeval *c)
{
	c->tv_sec = a->tv_sec - b->tv_sec;
	c->tv_usec = a->tv_usec - b->tv_usec;
	if (c->tv_usec < 0) {
		c->tv_sec -= 1;
		c->tv_usec += 1000000;
	}
}

int
rs_waitread(socket_t sd, unsigned long ms)
{
	struct timeval start, tv, tv2, cur;
	fd_set fds;
	ioret_t rv;

	tv.tv_sec = ms / 1000;
	tv.tv_usec = 1000 * (ms % 1000);

	/* select on Linux does not return the remaining time on
	   error, so we manually track it. */
	pi_gettimeofday(&start, NULL);
	memcpy(&tv2, &tv, sizeof(tv2));
retry:
	FD_ZERO(&fds);
	FD_SET(sd, &fds);
	rv = select((int)sd+1, &fds, NULL, NULL, &tv);
	if (rv > 0)
		return 0;  /* sd is ready */
	if (rv == 0)
		return -1; /* timeout */
	if (rv < 0 && errno != EINTR)
		return -1; /* error */

	/* We were interrupted.  Retry for remaining time. */
	pi_gettimeofday(&cur, NULL);
	rs_tv_diff(&cur, &start, &tv);
	rs_tv_diff(&tv2, &tv, &tv);
	if (tv.tv_sec < 0 || (0 == tv.tv_sec && 0 == tv.tv_usec))
		return -1; /* timeout */
	goto retry;
}

/* Perform a blocking read of SD until LEN bytes are read into BUF.
   If MS is non-zero, return after MS milliseconds, no matter how much
   has been read.  Return -1 on this timeout or other error. */
int
rs_xread(socket_t sd, iobuff_t buf, iosize_t len, unsigned long ms)
{
	char *p = (char *)buf;
	iosize_t nrecv = 0;
	ioret_t rv;
	int ret;

	assert(len > 0);
	ret = -1;
	while (nrecv < len) {
		if (ms && 0 > rs_waitread(sd, ms))
			goto out; /* timeout */
		rv = recv(sd, p, len - nrecv, 0);
		if (0 > rv && EINTR == errno)
			continue;
		if (0 > rv)
			goto out;
		if (rv == 0) {
			ret = 0; /* closed */
			goto out;
		}
		nrecv += rv;
		p += rv;
	}
	ret = nrecv;
out:
	
	return ret;
}

int
rs_reset_on_close(socket_t sd, int onoff)
{
	struct linger l;

	/* VERSION CHECK NOT DONE ON WINDOWS */
#if 0
	static int version = 0;
	/* Linux 2.2 does not support reset-on-close.  But if we try
	   to set it, we will put the socket in linger-on-close mode
	   at maximum timeout.
	   FIXME: This is not a good version check. */
	if (!version) {
		int rv;
		struct utsname uts;
		rv = uname(&uts);
		assert(!rv);
		if (!strncmp(uts.release, "2.4", 3)) {
			version = 24;
		} else {
			version = -1;
		}
	}
	if (24 != version)
		return 0;
#endif

	l.l_onoff = onoff;
	l.l_linger = 0;
	if (0 > setsockopt(sd, SOL_SOCKET, SO_LINGER,
	                   (iobuff_t)&l, sizeof(l)))
		return -1;
	return 0;
}

int
rs_reuseaddr(socket_t sd)
{
	int optval = 1;

	if (0 > setsockopt(sd, SOL_SOCKET, SO_REUSEADDR,
			   (iobuff_t)&optval, sizeof(optval)))
		return -1;
	return 0;
}

int
rs_nodelay(socket_t sd, int optval)
{
	if (0 > setsockopt(sd, SOL_TCP, TCP_NODELAY,
	                   (iobuff_t)&optval, sizeof(optval)))
		return -1;
	return 0;
}

int
rs_nonblock(socket_t sd, int on)
{
	return pi_nonblock(sd, on);
}

int
rs_settcpbuf(socket_t sd, int type, int size)
{
	if (type != SO_SNDBUF && type != SO_RCVBUF) {
		errno = EINVAL;
		return -1;
	}
	if (0 > setsockopt(sd, SOL_SOCKET, type,
	                   (iobuff_t)&size, sizeof(size))) {
		rs_log("Warning: can't set %s to %d bytes failed: %s",
		       type == SO_SNDBUF ? "SO_SNDBUF" : "SO_RCVBUF",
		       size, strerror(errno));
		return -1;
	}
	return 0;
}

char *
rs_ipstr(struct sockaddr_in *sa)
{
	static char buf[128];
	struct servent *servent;

	static char addr[32];
	static char serv[32];

	strcpy(addr, inet_ntoa(sa->sin_addr));
	servent = getservbyport(sa->sin_port, "tcp");
	if (servent) {
		strcpy(serv, servent->s_name);
		sprintf(buf, "%s:%s", addr, serv);
	}
	else
		sprintf(buf, "%s:%d", addr, ntohs(sa->sin_port));
	return buf;
}
