/*
 * rocks/util_unix.c
 *
 * UNIX-specific utility functions.
 */

#include "rs_unix.h"
#include "log.h"
#include <ctype.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

static const char xtoa[] = { '0', '1', '2', '3',
			     '4', '5', '6', '7',
			     '8', '9', 'a', 'b',
			     'c', 'd', 'e', 'f' };

/* BUF must be at least 58 bytes long */
static void
data_to_str(const char *data, int len, char *buf)
{
	int i;
	char *p = buf;
	char *q = buf + 39;
	const char *r = data;
	
	*q++ = ' ';
	*q++ = ' ';
	for (i = 0; i < 16 && i < len; i++) {
		if (i > 0 && (i % 2) == 0)
			*p++ = ' ';
		*p++ = xtoa[(*r >> 4) & 0xf];
		*p++ = xtoa[*r & 0xf];
		*q++ = (isprint(*r) ? *r : '.');
		r++;
	}

	for (; i < 16; i++) {
		if (i > 0 && (i % 2) == 0)
			*p++ = ' ';
		*p++ = '0';
		*p++ = '0';
		*q++ = '.';
	}
	*q = '\0';
}

void
rs_logbytes(char *bytes, int len)
{
	char buf[64];
	while (len > 0) {
		int nb = MIN(len, 16);
		data_to_str(bytes, nb, buf);
		rs_log("%s", buf);
		bytes += nb;
		len -= nb;
	}
}

int
rs_rocklist_insert(rocklist_t *head, rs_t rs)
{
	rocklist_t rl = (rocklist_t) malloc(sizeof(struct rocklist));
	if (!rl)
		return -1;
	rl->rs = rs;
	rl->next = *head;
	*head = rl;
	return 0;
}

int
rs_rocklist_remove(rocklist_t *head, rs_t rs)
{
	rocklist_t p, q;

	p = *head;
	if (!p)
		return -1; /* Not found */

	/* List head */
	if (p->rs == rs) {
		*head = p->next;
		free(p);
		return 0;
	}

	/* Rest of list */
	q = p;
	p = q->next;
	while (p) {
		if (p->rs == rs) {
			q->next = p->next;
			free(p);
			return 0;
		}
		q = p;
		p = q->next;
	}
	return -1; /* Not found */
}

rs_t
rs_rocklist_findsa(rocklist_t head, struct sockaddr_in *sa)
{
	rocklist_t p = head;
	while (p) {
		struct sockaddr_in *ta;
		ta = &p->rs->sa_peer;
		if (ta->sin_addr.s_addr == sa->sin_addr.s_addr
		    && ta->sin_port == sa->sin_port)
			return p->rs;
		p = p->next;
	}
	return NULL;
}

void
rs_rocklist_free(rocklist_t *head)
{
	rocklist_t p = *head;
	if (p)
		rs_rocklist_free(&p->next);
	*head = NULL;
}

void
rs_kill9_and_wait(pid_t pid)
{
	int olderrno;
	sigset_t mask, old;

	olderrno = errno;
	sigemptyset(&mask);
	sigaddset(&mask, SIGCHLD);
	if (0 > sigprocmask(SIG_BLOCK, &mask, &old))
		goto out;
	if (0 == kill(pid, SIGKILL) && 0 > waitpid(pid, NULL, 0))
		goto out;
	sigprocmask(SIG_SETMASK, &old, NULL);
out:
	errno = olderrno;
}

static int
parse_addr(const char *s, struct in_addr *addr)
{
	struct hostent* h;
	h = gethostbyname(s);
	if (!h)
		return -1;
	*addr = *((struct in_addr *) h->h_addr); /* network order */
	return 0;
}

static int
parse_port(const char *s, short *port)
{
	char *p;
	struct servent *se;
	unsigned long l;

	se = getservbyname(s, "tcp");
	if (se) {
		*port = se->s_port;
		return 0;
	}
	l = strtoul(s, &p, 10);
	if (*p != '\0')
		return -1;
	*port = (short) htons(l);
	return 0;
}

static int
parse_ip(const char *s, struct sockaddr_in *addr)
{
	char *buf = NULL;
	char *p;
	int ret = -1;

	buf = strdup(s);
	if (!buf) {
		fprintf(stderr, "Out of memory\n");
		return -1;
	}
	addr->sin_family = AF_INET;
	addr->sin_addr.s_addr = INADDR_ANY;
	addr->sin_port = htons(0);
	if ((p = strchr(buf, ':'))) {
		/* HOST:PORT */
		*p++ = '\0';
		if (0 > parse_addr(buf, &addr->sin_addr))
			goto out;
		if (0 > parse_port(p, &addr->sin_port))
			goto out;
	} else if ((p = strchr(buf, '.'))) {
		/* HOST */
		if (0 > parse_addr(buf, &addr->sin_addr))
			goto out;
	} else {
		/* PORT or HOST? */
		if (0 > parse_port(buf, &addr->sin_port)
		    && 0 > parse_addr(buf, &addr->sin_addr))
			goto out;
	}
	ret = 0;
out:
	free(buf);
	return ret;
}
